__author__ = 'robert'





import inspect
import pypet.petexceptions as pex
from pypet import globally
from pypet.annotations import WithAnnotations
import logging


#For fetching:
STORE = 'STORE'
LOAD = 'LOAD' #We want to f_load stuff with the storage service
REMOVE = 'REMOVE' #We want to remove stuff


#Group Constants
RESULT = 'RESULT'
RESULT_GROUP = 'RESULTGROUP'
PARAMETER = 'PARAMETER'
PARAMETER_GROUP = 'PARAMETER_GROUP'
DERIVED_PARAMETER = 'DERIVED_PARAMETER'
DERIVED_PARAMETER_GROUP = 'DERIVED_PARAMETER_GROUP'
CONFIG = 'CONFIG'
CONFIG_GROUP = 'CONFIG_GROUP'

## For fast searching
FAST_UPPER_BOUND = 2


#### Naming
FORMAT_ZEROS=8
RUN_NAME = 'run_'
FORMATTED_RUN_NAME=RUN_NAME+'%0'+str(FORMAT_ZEROS)+'d'

class NNTreeNode(WithAnnotations):
    def __init__(self,full_name,root,leaf):
        #self.v_full_name = full_name

        super(NNTreeNode,self).__init__()

        split_name = full_name.split('.')

        if root:
            self._depth=0
        else:
            self._depth = len(split_name)

        self._name=split_name.pop()
        self._location='.'.join(split_name)
        self._leaf=leaf

    @property
    def v_depth(self):
        ''' Depth of the node in the trajectory tree.
        '''
        return self._depth

    @property
    def v_leaf(self):
        '''Whether node is a leaf or not (i.e. it is a group node)'''
        return self._leaf



    def f_is_root(self):
        '''Returns whether the group is root (True for the trajectory and a single run object)'''
        return self._depth==0

    @property
    def v_full_name(self):
        ''' The full name, relative to the root node.

        The full name of a trajectory or single run is the empty string since it is root.
        '''
        if self.f_is_root():
            return ''
        else:
            if self._location=='':
                return self._name
            else:
                return '.'.join([self._location,self._name])

    @property
    def v_name(self):
        ''' Name of the node
        '''
        return self._name

    @property
    def v_location(self):
        ''' Location relative to the root node.

        The location of a trajectory or single run is the empty string since it is root.
        '''
        return self._location

    def _rename(self, fullname):
        ''' Renames the parameter or result.
        '''
        #self.v_full_name = fullname
        split_name = fullname.split('.')

        self._depth = len(split_name)

        self._name=split_name.pop()
        self._location='.'.join(split_name)


    def f_get_class_name(self):
        ''' Returns the class name of the parameter or result or group,
        equivalent to `return self.__class__.__name__`
        '''
        return self.__class__.__name__



class NNLeafNode(NNTreeNode):
    ''' Abstract class interface of result or parameter
    '''

    def __init__(self,full_name,comment,parameter):
        super(NNLeafNode,self).__init__(full_name=full_name,root=False,leaf=True)
        self._parameter=parameter
        self._fast_accessible = parameter

        self._comment=''
        self.v_comment=comment


    @property
    def v_comment(self):
        ''' Should be a nice descriptive comment'''
        return self._comment

    @v_comment.setter
    def v_comment(self, comment):

        comment = str(comment)
        if len(comment)>=globally.HDF5_STRCOL_MAX_COMMENT_LENGTH:
            raise AttributeError('Your comment is too long ( %d characters), only comments up to'
                                 '%d characters are allowed.' %
                                 (len(comment),globally.HDF5_STRCOL_MAX_COMMENT_LENGTH))

        self._comment=comment


    @property
    def v_fast_accessible(self):
        '''Whether or not fast access can be supported by the Parameter or Result'''
        return self._fast_accessible

    @property
    def v_parameter(self):
        '''Whether the node is a parameter or not (i.e. a result)'''
        return self._parameter

    def f_val_to_str(self):
        ''' Returns a string summarizing the data handled by the parameter or result
        '''
        return ''


    def __str__(self):
        ''' String representation of the parameter or result, If not other specified this is
        simply the full name.
        '''
        return self.v_full_name





    def _store_flags(self):
        ''' We do not return storage flags, we let the storage service infer how to store
        stuff from the data itself.

        If you write your own parameter you can make specifications on how to store data,
        see also :func:`pypet.storageservice.HDF5StorageService.store`.

        '''
        return {}

    def _store(self):
        ''' Method called by the storage service for serialization.

        The method converts the parameter's or result's value into  simple
        data structures that can be stored to disk.
        Returns a dictionary containing these simple structures.
        Understood basic structures are

        * python natives (int,str,bool,float,complex)

        * python lists and tuples

        * numpy natives and arrays of type np.int8-64, np.uint8-64, np.float32-64,
                                            np.complex, np.str

        * python dictionaries of the previous types (flat not nested!)

        * pandas data frames

        * object tables

        :return: A dictionary containing basic data structures.

        '''
        raise NotImplementedError('Implement this!')

    def _load(self, load_dict):
        ''' Method called by the storage service to reconstruct the original result.

        Data contained in the load_dict is equal to the data provided by the result
        when previously called with _store()

        :param load_dict: The dictionary containing basic data structures

        '''
        raise  NotImplementedError('Implement this!')


    def f_is_empty(self):
        ''' Returns true if no data is handled by a result or parameter
        '''
        raise NotImplementedError('You should implement this!')

    def f_empty(self):
        ''' Removes all data from the result.

        If the result has already been stored to disk via a trajectory and a storage service,
        the data on disk is not affected by `f_emptyty`.

        Yet, this function is particularly useful if you have stored very large data to disk
        and you want to free some memory on RAM but still keep the skeleton of the result.

        '''
        raise NotImplementedError('You should implement this!')

class NaturalNamingInterface(object):


    # TYPE_NODE_MAPPING = {
    #     'result_group' : ResultGroup,
    #     'config_group' : ConfigGroup,
    #     'derived_parameter_group' : DerivedParameterGroup
    #     'parameter_group' : ParameterGroup
    # }




    def __init__(self, root_instance):

        self._root_instance = root_instance

        self._run_or_traj_name= root_instance.v_name

        self._logger = logging.getLogger('pypet.trajectory.NNTree=' +
                                         self._root_instance.v_name)

        self._flat_storage_dict = {}

        self._nodes_and_leaves = {}

        self._not_admissible_names = set(dir(self)) | set( dir(self._root_instance) )


    def _map_type_to_dict(self,type_name):

        root = self._root_instance

        if type_name == RESULT:
            return root._results
        elif type_name == PARAMETER:
            return root._parameters
        elif type_name == DERIVED_PARAMETER:
            return root._derived_parameters
        elif type_name == CONFIG:
            return root._config
        else:
            raise RuntimeError('You shall not pass!')


    def _change_root(self, new_root):
        new_root._children = self._root_instance._children
        self._root_instance = new_root
        self._run_or_traj_name= self._root_instance.v_name

        self._logger = logging.getLogger('pypet.trajectory.NNTree=' +
                                         self._root_instance.v_name)

    def _get_fast_access(self):
        return self._root_instance.v_fast_access

    def _get_search_strategy(self):
        return self._root_instance.v_search_strategy

    def _get_check_uniqueness(self):
        return self._root_instance.v_check_uniqueness


    def _fetch_from_string(self,store_load, name, args, kwargs):

        node = self._root_instance.f_get(name)

        return self._fetch_from_node(store_load,node,args,kwargs)

    def _fetch_from_node(self,store_load, node,args,kwargs):
        msg = self._node_to_msg(store_load,node)

        return (msg,node,args,kwargs)

    def _fetch_from_tuple(self,store_load,store_tuple,args,kwargs):

        node = store_tuple[1]
        msg = store_tuple[0]
        if len(store_tuple) > 2:
            args = store_tuple[2]
        if len(store_tuple) > 3:
            kwargs = store_tuple[3]
        if len(store_tuple) > 4:
            print store_tuple
            raise TypeError('Your argument tuple has to many entries, please call stuff '
                            'to f_store with [(msg,item,args,kwargs),]')

        ##dummy test
        _ = self._fetch_from_node(store_load,node)

        return (msg,node,args,kwargs)


    @staticmethod
    def _node_to_msg(store_load,node):

        if node.v_leaf:
            if store_load ==STORE:
                return globally.LEAF
            elif store_load == LOAD:
                return globally.LEAF
            elif store_load == REMOVE:
                 return globally.REMOVE
        else:
            if store_load ==STORE:
                return globally.GROUP
            elif store_load == LOAD:
                return globally.GROUP
            elif store_load == REMOVE:
                 return globally.REMOVE
        #raise ValueError('I do not know how to f_store >>%s<<' % str(node))


    def _fetch_items(self, store_load, iterable, args, kwargs):
        only_empties = kwargs.pop('only_empties', False)

        non_empties = kwargs.pop('non_empties', False)


        item_list = []
        for iter_item in iterable:

            try:
                item_tuple = self._fetch_from_string(store_load,iter_item,args,kwargs)
            except AttributeError:
                try:
                    item_tuple = self._fetch_from_node(store_load,iter_item,args,kwargs)
                except AttributeError:
                    item_tuple = self._fetch_from_tuple(store_load,iter_item,args,kwargs)

            item = item_tuple[1]
            msg = item_tuple[0]

            if only_empties and not item.f_is_empty():
                continue
            if non_empties and item.f_is_empty():
                continue

            # if self._root_instance._is_run:
            #     fullname = item.v_full_name
            #     if not self._root_instance.v_name in fullname:
            #         raise TypeError('You want to store/load >>%s<< but this belongs to the '
            #                         'parent trajectory not to the current single run.' % fullname)

            if (msg == globally.REMOVE and
                        item.v_full_name in self._root_instance._explored_parameters):
                raise TypeError('You cannot remove an explored parameter of a trajectory stored '
                                'into an hdf5 file.')


            item_list.append(item_tuple)
        return item_list


    def _remove_subtree(self,start_node,name):

        def _remove_subtree_inner(node):
            if not node.v_leaf:

                for name in node._children.keys():
                    child = node._children[name]
                    _remove_subtree_inner(child)
                    del node._children[name]
                    del child

            self._delete_node(node)



        child = start_node._children[name]

        _remove_subtree_inner(child)
        del start_node._children[name]






    def _delete_node(self,node):

        full_name = node.v_full_name
        name = node.v_name

        root = self._root_instance


        if full_name in ['parameters','results','derived_parameters','config','']:
            ## You cannot delete root or nodes in first layer
            return

        if node.v_leaf:
            if full_name in root._parameters:
                del root._parameters[full_name]

            elif full_name in root._config:
                del root._config[full_name]

            elif full_name in root._derived_parameters:
                del root._derived_parameters[full_name]

            elif full_name in root._results:
                del root._results[full_name]

            if full_name in root._explored_parameters:
                del root._explored_parameters[full_name]

                if len(root._explored_parameters)==0:
                    if root._stored:
                       self._logger.warning('You removed an explored parameter, but your '
                                            'trajectory was already stored to disk. So it is '
                                            'not shrunk!')
                    else:
                        root.f_shrink()

            del self._flat_storage_dict[full_name]



        else:
            del root._groups[full_name]

        del self._nodes_and_leaves[name][full_name]
        if len(self._nodes_and_leaves[name]) == 0:
            del self._nodes_and_leaves[name]


    def _remove_node_or_leaf(self, instance,remove_empty_groups):
        full_name = instance.v_full_name
        split_name = full_name.split('.')
        self._remove_recursive(self._root_instance,split_name,remove_empty_groups)



    def _remove_recursive(self,actual_node,split_name,remove_empty_groups):

        if len(split_name)== 0:
            self._delete_node(actual_node)
            return True

        name = split_name.pop(0)


        child = actual_node._children[name]

        if self._remove_recursive(child, split_name, remove_empty_groups):
            del actual_node._children[name]
            del child

            if remove_empty_groups and len(actual_node._children) == 0:
                self._delete_node(actual_node)
                return True

        return False





    def _shortcut(self, name):

        expanded = None

        if name.startswith('run_') or name.startswith('r_'):
            split_name = name.split('_')
            if len(split_name) == 2:
                index = split_name[1]
                if index.isdigit():
                    if len(index) < FORMAT_ZEROS:
                        expanded = FORMATTED_RUN_NAME % int(index)

        if name in ['cr', 'currentrun', 'current_run']:
            expanded = self._root_instance.v_name

        if name in ['par', 'param']:
            expanded = 'parameters'

        if name in ['dpar', 'dparam']:
            expanded = 'derived_parameters'

        if name in ['res']:
            expanded = 'results'

        if name in ['conf']:
            expanded = 'config'

        if name in ['traj', 'tr']:
            expanded = 'trajectory'

        return expanded


    def _add_prefix( self,name, start_node, group_type_name):

        ## If we add an instance with a full name at root, we do not need to add the prefix
        if start_node.f_is_root():
            for prefix in ('results','parameters','derived_parameters','config'):
                if name.startswith(prefix):
                    return name

        root=self._root_instance

        if start_node.v_depth<2:
            if group_type_name == DERIVED_PARAMETER_GROUP:

                add=''

                if not ((name.startswith('run_') and len(name) ==len(RUN_NAME)+FORMAT_ZEROS) or
                    name.startswith('trajectory')):

                    if root._is_run:
                        add= start_node.v_name + '.'
                    else:
                        add= 'trajectory.'

                if start_node.v_depth== 0:
                    add = 'derived_parameters.' + add

                return add+name


            elif group_type_name == RESULT_GROUP:

                add = ''

                if not ((name.startswith('run_') and len(name) ==len(RUN_NAME)+FORMAT_ZEROS) or
                    name == 'trajectory'):

                    if root._is_run:
                        add= start_node.v_name + '.'
                    else:
                        add= 'trajectory.'

                if start_node.v_depth== 0:
                    add = 'results.' + add

                return add+name

        if start_node.f_is_root():

            if group_type_name == PARAMETER_GROUP:
                return 'parameters.'+name

            if group_type_name == CONFIG_GROUP:
                return 'config.'+name

        return name


    def _add_from_leaf_instance(self,start_node,instance):

        full_name = start_node.v_full_name

        if full_name.startswith('results'):
            group_type_name = RESULT_GROUP
            type_name = RESULT


        elif full_name.startswith('parameters'):
            group_type_name = PARAMETER_GROUP
            type_name = PARAMETER

        elif full_name.startswith('derived_parameters'):
            group_type_name = DERIVED_PARAMETER_GROUP
            type_name = DERIVED_PARAMETER

        elif full_name.startswith('config'):
            group_type_name = CONFIG_GROUP
            type_name = CONFIG

        else:
            raise RuntimeError('You shall not pass!')

        return self._add_generic(start_node,type_name,group_type_name,[instance],{})


    def _add_from_group_name(self,start_node,name):

        full_name = start_node.v_full_name

        if full_name.startswith('results'):
            group_type_name = RESULT_GROUP


        elif full_name.startswith('parameters'):
            group_type_name = PARAMETER_GROUP


        elif full_name.startswith('derived_parameters'):
            group_type_name = DERIVED_PARAMETER_GROUP


        elif full_name.startswith('config'):
            group_type_name = CONFIG_GROUP


        else:
            raise RuntimeError('You shall not pass!')

        return self._add_generic(start_node,group_type_name,group_type_name,[name],{})



    def _add_generic(self,start_node,type_name,group_type_name,args,kwargs):

        if type_name == group_type_name:
            # Wee add a group node, this can be only done by name:
            name = args[0]
            instance = None
            constructor = None

        else:
            ## We add a leaf node in the end:
            args = list(args)

            create_new = True

            if len(args) == 1 and len(kwargs)==0:
                item = args[0]
                try:
                    name = item.v_full_name
                    instance = item
                    constructor = None

                    create_new = False
                except AttributeError:
                    pass




            if create_new:
                if inspect.isclass(args[0]):
                    constructor = args.pop(0)
                else:
                    constructor = None

                instance = None

                name = args.pop(0)

        name = self._add_prefix(name,start_node,group_type_name)


        return self._add_to_nninterface(start_node,name,type_name,group_type_name,instance,
                                            constructor,args,kwargs)








    def _add_to_nninterface(self, start_node, name, type_name, group_type_name,
                            instance, constructor, args, kwargs ):


        split_name = name.split('.')

        faulty_names= self._check_names(split_name)

        if faulty_names:
            raise AttributeError(
                'Your Parameter/Result/Node >>%s<< f_contains the following not admissible names: '
                '%s please choose other names.'
                % (name, faulty_names))


        return self._add_to_tree(start_node, split_name, type_name, group_type_name, instance, constructor,
                     args, kwargs )


    def _add_to_tree(self, start_node, split_name, type_name, group_type_name, instance, constructor,
                     args, kwargs ):
        try:
            act_node = start_node
            last_idx = len(split_name)-1
            for idx, name in enumerate(split_name):

                if not name in act_node._children:


                    if idx == last_idx and group_type_name != type_name:
                        # If we are at the end of the chain and we add a leave node

                        new_node = self._create_any_param_or_result(act_node.v_full_name,name,
                                                                    type_name,instance,constructor,
                                                                    args,kwargs)

                        self._flat_storage_dict[new_node.v_full_name] = new_node


                    else:
                        # We add a group node
                        new_node = self._create_any_group(act_node.v_full_name,name,
                                                          group_type_name)

                    act_node._children[name] = new_node


                    if not name in self._nodes_and_leaves:
                        self._nodes_and_leaves[name]={new_node.v_full_name:new_node}
                    else:
                        self._nodes_and_leaves[name][new_node.v_full_name]=new_node

                else:
                    if idx == last_idx:
                        raise AttributeError('You already have a group/instance >>%s<< under '
                                             '>>%s<<' % (name,start_node.v_full_name))


                act_node = act_node._children[name]

            return act_node
        except:
            self._logger.error('Failed storing >>%s<< under >>%s<<.' %
                               (name, start_node.v_full_name))
            raise



    def _check_names(self, split_names):

        faulty_names = ''



        for split_name in split_names:
            if split_name in self._not_admissible_names:
                faulty_names = '%s %s is a method/attribute of the trajectory/treenode/naminginterface,' % \
                               (faulty_names, split_name)

            if split_name[0] == '_':
                faulty_names = '%s %s starts with a leading underscore,' % (
                    faulty_names, split_name)

            if ' ' in split_name:
                faulty_names = '%s %s f_contains white space(s),' % (faulty_names, split_name)

            if not self._shortcut(split_name) is None:
                faulty_names = '%s %s is already an important shortcut,' % (faulty_names, split_name)


        name = split_names.pop()
        location = '.'.join(split_names)
        if len(name)>= globally.HDF5_STRCOL_MAX_NAME_LENGTH:
            faulty_names = '%s %s is too long the name can only have %d characters but it has %d,' % \
                           (faulty_names,name,len(name),globally.HDF5_STRCOL_MAX_NAME_LENGTH)

        if len(location) >= globally.HDF5_STRCOL_MAX_LOCATION_LENGTH:
            faulty_names = '%s %s is too long the location can only have %d characters but it has %d,' % \
                           (faulty_names,name,len(location),globally.HDF5_STRCOL_MAX_LOCATION_LENGTH)

        split_names.append(name)
        return faulty_names


    def _create_any_group(self, location, name, type_name):

        instance = self._create_node_instance(location,name, type_name)

        self._root_instance._groups[instance.v_full_name]=instance


        return instance



    def _create_node_instance(self,location,name,type_name):

        if location:
            full_name = '%s.%s' % (location, name)
        else:
            full_name = name


        if type_name == RESULT_GROUP:
            return ResultGroup(self,full_name=full_name, root=False)

        elif type_name == PARAMETER_GROUP:
            return ParameterGroup(self,full_name=full_name, root=False)

        elif type_name == CONFIG_GROUP:
            return ConfigGroup(self,full_name=full_name, root=False)

        elif type_name == DERIVED_PARAMETER_GROUP:
            return DerivedParameterGroup(self,full_name=full_name, root=False)
        else:
            raise RuntimeError('You shall not pass!')


    def _create_any_param_or_result(self, location, name, type_name, instance, constructor,
                                    args, kwargs):


        root = self._root_instance

        if location:
            full_name = '%s.%s' % (location, name)
        else:
            full_name = name

        if instance is None:

            if constructor is None:
                if type_name == RESULT:
                    constructor=root._standard_result
                else:
                    constructor=root._standard_parameter



            instance = constructor(full_name, *args, **kwargs)
        else:
            instance._rename(full_name)

        
        where_dict = self._map_type_to_dict(type_name)

        if full_name in where_dict:
            raise AttributeError(full_name + ' is already part of trajectory,')

        if type_name != RESULT:
            if full_name in root._changed_default_parameters:
                self._logger.info('You have marked parameter %s for change before, so here you go!' %
                                  full_name)

                change_args, change_kwargs = root._changed_default_parameters.pop(full_name)
                instance.f_set(*change_args, **change_kwargs)

        where_dict[full_name] = instance




        self._logger.debug('Added >>%s<< to trajectory.' % full_name)

        return instance







    def __getstate__(self):
        result = self.__dict__.copy()
        del result['_logger']
        return result


    def __setstate__(self, statedict):
        self.__dict__.update(statedict)
        self._logger = logging.getLogger('pypet.trajectory.NNTree=' +
                                         self._run_or_traj_name)


    @staticmethod
    def _get_result(data, fast_access):

        if fast_access and data.v_fast_accessible:
            return data.f_get()
        else:
            return data

    @staticmethod
    def _iter_nodes( node, recursive=False, search_strategy=globally.BFS):

        if recursive:
            if search_strategy == globally.BFS:
                return NaturalNamingInterface._recursive_traversal_bfs(node)
            elif search_strategy == globally.DFS:
                return NaturalNamingInterface._recursive_traversal_dfs(node)
            else:
                raise ValueError('Your search method is not understood!')
        else:
            return node._children.itervalues()


    @staticmethod
    def _iter_leaves(node):
        for node in node.f_iter_nodes(recursive=True):
            if node.v_leaf:
                yield node
            else:
                continue

    def _to_dict(self,node,fast_access = True, short_names=False, copy=True):

        if (fast_access or short_names)  and not copy:
            raise ValueError('You can not request the original data with >>fast_access=True<< or'
                             ' >>short_names=True<<.')

        if node.f_is_root():
            temp_dict = self._flat_storage_dict

            if not fast_access and not short_names:
                if copy:
                    return temp_dict.copy()
                else:
                    return temp_dict

            else:
                iterator = temp_dict.itervalues()
        else:
            iterator=self._iter_leaves(node)

        result_dict={}
        for val in iterator:
            if short_names:
                new_key = val.v_name
            else:
                new_key = val.v_full_name

            if new_key in result_dict:
                raise ValueError('Your short names are not unique!')

            new_val = self._get_result(val,fast_access)
            result_dict[new_key]=new_val

        return result_dict

    @staticmethod
    def _recursive_traversal_bfs(node):
        if not node._leaf:
            for child in node._children.itervalues():
                yield child

            for child in node._children.itervalues():
                for new_node in NaturalNamingInterface._recursive_traversal_bfs(child):
                    yield new_node


    @staticmethod
    def _recursive_traversal_dfs(node):
        if not node._leaf:
            for child in node._children.itervalues():
                yield child
                for new_node in NaturalNamingInterface._recursive_traversal_dfs(child):
                    yield new_node


    def _very_fast_search(self, node, key, check_uniqueness):
        ''' Will always test if nodes are not unique'''

        parent_full_name = node.v_full_name

        candidate_dict = self._nodes_and_leaves[key]

        if len(candidate_dict) > FAST_UPPER_BOUND:
            raise pex.TooManyGroupsError('Too many nodes')

        result_node = None
        for goal_name in candidate_dict.iterkeys():

            if goal_name.startswith(parent_full_name):

                if not result_node is None:
                    raise pex.NotUniqueNodeError('Node >>%s<< has been found more than once,'
                                                 'full name of first found is >>%s<< and of'
                                                 'second >>%s<<'
                                                 % (key,goal_name,result_node.v_full_name))

                result_node=candidate_dict[goal_name]

        return result_node


    def _search(self,node,  key, check_uniqueness, search_strategy):

        try:
            return self._very_fast_search(node, key, check_uniqueness)
        except pex.TooManyGroupsError:
            pass
        except pex.NotUniqueNodeError:
            if check_uniqueness:
                raise
            else:
                pass

        nodes_iterator = NaturalNamingInterface._iter_nodes(node, recursive=True,
                                                            search_strategy=search_strategy)

        result_node = None
        for child in nodes_iterator:
            if key == child.v_name:

                if not result_node is None:
                    raise pex.NotUniqueNodeError('Node >>%s<< has been found more than once,'
                                                 'full name of first found is >>%s<< and of '
                                                 'second >>%s<<'
                                                 % (key,child.v_full_name,result_node.v_full_name))

                result_node =  child
                if not check_uniqueness:
                    return result_node

        return result_node



    def _get(self,node, name, fast_access, check_uniqueness, search_strategy):
        ''' Same as traj.>>name<<
        Requesting parameters via f_get does not pay attention to fast access. Whether the parameter object or it's
        default evaluation is returned depends on the value of >>fast_access<<.

        :param name: The Name of the Parameter,Result or NNGroupNode that is requested.
        :param fast_access: If the default evaluation of a parameter should be returned.
        :param check_uniqueness: If search through the Parameter tree should be stopped after finding an entry or
        whether it should be chekced if the path through the tree is not unique.
        :param: search_strategy: The strategy to search the tree, either breadth first search (globally.BFS) or depth first
        seach (globally.DFS).
        :return: The requested object or it's default evaluation. Raises an error if the object cannot be found.
        '''


        split_name = name.split('.')


        ## Rename shortcuts and check keys:
        for idx, key in enumerate(split_name):
            shortcut = self._shortcut(key)
            if shortcut:
                key = shortcut
                split_name[idx] = key

            if key[0] == '_':
                raise AttributeError('Leading underscores are not allowed for group or parameter '
                                     'names. Cannot return %s.' % key)

            if not key in self._nodes_and_leaves:
                raise AttributeError('%s is not part of your trajectory or it\'s tree.' % name)

        ## Check in O(1) if len(split_name)==1
        if len(split_name)==1 and not check_uniqueness:
            result = node._children.get(key,None)
        else:
            result = None

        if result is None:
            ## Check in O(d) first if a full parameter/result name is given
            fullname = '.'.join(split_name)

            if fullname.startswith(node.v_full_name):
                new_name = fullname
            else:
                new_name = node.v_full_name + '.' + fullname

            if new_name in self._flat_storage_dict:
                return self._get_result(self._flat_storage_dict[new_name],
                                                     fast_access=fast_access)

            if new_name in self._root_instance._groups:
                return self._root_instance._groups[new_name]

        # Check in O(N)
        # [Worst Case, Average Case is better since looking into a single dict costs O(1)]
        # globally.BFS or globally.DFS
        # If check Uniqueness == True, search is slower since the full dictionary
        # is always searched

        result = node
        for key in split_name:
            result = self._search(result,key, check_uniqueness, search_strategy)

        if result is None:
            raise AttributeError('The node or param/result >>%s<<, cannot be found.' % name)
        if result.v_leaf:
            return self._get_result(result, fast_access)
        else:
            return result






class NNGroupNode(NNTreeNode):
    ''' A group node hanging somewhere under the trajectory or single run root node.
    You can add other groups or parameters/results to it.
    '''

    def __init__(self, nn_interface=None, full_name='', root=False):
        super(NNGroupNode,self).__init__(full_name,root=root,leaf=False)
        self._children={}
        self._nn_interface=nn_interface

    def __str__(self):

        if not self.f_is_root():
            name = self.v_full_name
        else:
            name = self.v_name

        return '<%s>: %s: %s' % (self.f_get_class_name(),name,
                                 str([(key,str(type(val)))
                                      for key,val in self._children.iteritems()]))

    def f_children(self):
        '''Returns the number of children of the group.'''
        return len(self._children)

    def f_has_children(self):
        '''Checks if node has children or not'''
        return len(self._children)==0

    def __getattr__(self, name):

        if (not '_nn_interface' in self.__dict__ or
                not '_full_name' in self.__dict__ or
                not '_children' in self.__dict__ or
                not 'f_get' in self.__class__.__dict__ or
                    name[0] == '_'):
            raise AttributeError('Wrong attribute %s. (And you this statement prevents pickling '
                                 'problems)' % name)

        return self.f_get(name, self._nn_interface._get_fast_access(),
                        self._nn_interface._get_check_uniqueness(),
                        self._nn_interface._get_search_strategy())

    def __contains__(self, item):
        return self.f_contains(item)

    def f_remove_child(self, name, recursive = False):
        ''' Removes a child of the group.

        Note that groups and leaves are only removed from the current trajectory in RAM.
        If the trajectory is stored to disk, this data is not affected. Thus, remove can be used
        only to free RAM memory!

        If you want to free memory on disk via your storage service,
        use :func:`~pypet.trajectory.Trajectory.f_remove_item(s)` of your trajectory.

        :param name: Name of child

        :param recursive: Must be true if child is a group that has children. Will remove
                            the whole subtree in this case. Otherwise a Type Error is thrown.

        :raises: TypeError

        '''
        if not name in self._children:
                raise TypeError('Your group >>%s<< does not contain the child >>%s<<.' %
                                (self.v_full_name,name))

        else:
            child = self._children[name]

            if not child.v_leaf and child.f_has_children and not recursive:
                raise TypeError('Cannot remove child. It is a group with children. Use'
                                ' f_remove with >>recursive = True')
            else:
                self._nn_interface._remove_subtree(self,name)


    def f_contains(self, item, recursive = True):
        ''' Checks if the node contains a specific parameter or result.

        It is checked if the item can be found via the "~pypet.naturalnaming.NNGroupNode.f_get" method.

        :param item: Parameter/Result name or instance.

        :param: recursive: Whether the whole sub tree under the group should be checked or only
                          its immediate children. Default is the whole sub tree.
                          If `recursive=False` you must only specify the name not the full name.

        :return: True or False
        '''

        try:
            search_string = item.v_full_name
            name = item.v_name
        except AttributeError:
            search_string = item
            name = item

        try:
            if search_string.startswith(self.v_full_name) and self.v_full_name != '':
                _, _, search_string = search_string.partition(self.v_full_name)
        except AttributeError:
            return False

        if recursive:
            try:
                self.f_get(search_string)

                return True
            except AttributeError:

                return False
        else:
            return name in self._children

    def __setattr__(self, key, value):
        if key.startswith('_'):
            self.__dict__[key] = value
        elif hasattr(self.__class__,key):
            property = getattr(self.__class__,key)
            if property.fset is None:
                raise AttributeError('%s is read only!' % key)
            else:
                property.fset(self,value)
        else:
            instance = self.f_get(key)

            if not instance.v_parameter:
                raise AttributeError('You cannot assign values to a tree node or a list of nodes '
                                     'and results, it only works for parameters ')

            instance.f_set(value)

    def __getattr__(self, key):
        if key.startswith('_'):
            raise AttributeError('Trajectory node does not contain >>%s<<' % key)

        if not '_nn_interface' in self.__dict__:
            raise AttributeError('This is to avoid pickling issues')

        return self._nn_interface._get(self,key,fast_access=self._nn_interface._get_fast_access(),
                                       check_uniqueness=self._nn_interface._get_check_uniqueness(),
                                       search_strategy=self._nn_interface._get_search_strategy())

    def f_iter_nodes(self, recursive=False, search_strategy=globally.BFS):
        ''' Iterates over nodes hanging below this group.

        :param recursive: Whether to iterate the whole sub tree or only immediate children.

        :param search_strategy: Either BFS or DFS (BFS recommended)

        :return: Iterator over nodes

        '''
        return self._nn_interface._iter_nodes(self,recursive=recursive,
                                             search_strategy=search_strategy)


    def f_iter_leaves(self):
        ''' Iterates (recursively) over all leaves hanging below the current group.
        '''
        return self._nn_interface._iter_leaves(self)


    def f_get(self, name, fast_access=False, check_uniqueness=False, search_strategy=globally.BFS):
        ''' Searches for an item (parameter/result/group node) with the given `name`-

        :param name: Name of the item (full name or parts of the full name)

        :param fast_access: If the result is a parameter, whether fast access should be applied.

        :param check_uniqueness: Whether it should be checked if the name unambiguously yields
                                 a single result.

        :param search_strategy: The search strategy (default and recommended is BFS)

        :return: The found instance (result/parameter/group node) or if fast access is True and you
                 found
                 a parameter, the parameter's value is returned.
        '''
        return self._nn_interface._get(self, name, fast_access=fast_access,
                                       check_uniqueness=check_uniqueness,
                                       search_strategy=search_strategy)

    def f_get_children(self, copy=True):
        '''Returns a children dictionary.

        :param copy: Whether the group's original dicitionary or a shallow copy is returned.
                     If you want the real dictionary please do not modify it at all!

        :returns: Dictionary of nodes
        '''
        if copy:
            return self._children.copy()
        else:
            return self._children

    def f_to_dict(self,fast_access = False, short_names=False):
        ''' Returns a dictionary with pairings of (full) names as keys and instances/values.

        :param fast_access: If True, parameter values are returned instead of the instances-

        :param short_names: If true, keys are not full names but only the names. Raises a Value
                            Error if the names are not unique.

        :return: dictionary

        :raises: ValueError

        '''
        return self._nn_interface._to_dict(self, fast_access=fast_access,short_names=short_names)


    def f_store_child(self,name,recursive=False):
        '''Stores a child or recursively a subtree to disk'''
        if not name in self._children:
                raise TypeError('Your group >>%s<< does not contain the child >>%s<<.' %
                                (self.v_full_name,name))

        traj = self._nn_interface._root_instance
        storage_service = traj.v_storage_service

        storage_service.store(globally.TREE, self._children[name], trajectory_name=traj.v_name,
                              recursive=recursive)



    def f_load_child(self,name,recursive=False,load_data=globally.UPDATE_DATA):
        '''Loads a child or recursively a subtree from disk.

        For how to choose 'load_data' see :ref:'loading'.
        '''
        if not name in self._children:
                raise TypeError('Your group >>%s<< does not contain the child >>%s<<.' %
                                (self.v_full_name,name))

        traj = self._nn_interface._root_instance
        storage_service = traj.v_storage_service

        storage_service.load(globally.TREE, self,child_name=name, trajectory_name=traj.v_name,
                             recursive=recursive, load_data=load_data, trajectory=traj)





class ParameterGroup(NNGroupNode):
    ''' Group node in your trajectory, hanging below `traj.parameters` (or the `parameters` group itself).

    You can add other groups or parameters to it.
    '''
    def f_add_parameter_group(self,name):
        '''Adds an empty parameter group under the current node.

        Adds the full name of the current node
        as prefix to the name of the group.
        If current node is the trajectory or single run (root), the prefix `'parameters'`
        is added to the full name.

        The `name` can also contain subgroups separated via colons, for example:
        `name=subgroup1.subgroup2.subgroup3`. These other parent groups will be automatically
        be created.

        '''
        return self._nn_interface._add_generic(self,type_name = PARAMETER_GROUP,
                                               group_type_name = PARAMETER_GROUP,
                                               args = (name,), kwargs={})

    def f_add_parameter(self,*args,**kwargs):
        ''' Adds a parameter under the current node.

        There are two ways to add a new parameter either by adding a parameter instance,
        directly:

        >>> new_parameter = Parameter('group1.group2.myparam', data=42, comment='Example!')
        >>> traj.f_add_parameter(new_parameter)

        Or by passing the values directly to the function, with the name being the first
        (non-keyword!) argument:

        >>> traj.f_add_parameter('group1.group2.myparam', data=42, comment='Example!')

        If you want to create a different parameter than the standard parameter, you can
        give the constructor as the first (non-keyword!) argument followed by the name (non-keyword!):

        >>> traj.f_add_parameter(PickleParameter,'group1.group2.myparam', data=42, comment='Example!')

        The full name of the current node is added as a prefix to the given parameter name.
        If the current node is the trajectory the prefix `'parameters'` is added to the name.

        '''
        return self._nn_interface._add_generic(self,type_name = PARAMETER,
                                               group_type_name = PARAMETER_GROUP,
                                               args=args,kwargs=kwargs)

class ResultGroup(NNGroupNode):
    ''' Group node in your trajectory, hanging below `traj.results`.

    You can add other groups or results to it.
    '''
    def f_add_result_group(self,name):
        '''Adds an empty result group under the current node.

        Adds the full name of the current node
        as prefix to the name of the group.
        If current node is the trajectory (root) adds the prefix `'results.trajectory'` to the full name.
        If current node is a single run (root) adds the prefix `'results.run_08%d%'` to the full name
        where `'08%d'` is replaced by the index of the current run.

        The `name` can also contain subgroups separated via colons, for example:
        `name=subgroup1.subgroup2.subgroup3`. These other parent groups will be automatically
        be created.

        '''


        return self._nn_interface._add_generic(self,type_name = RESULT_GROUP,
                                               group_type_name = RESULT_GROUP,
                                               args = (name,), kwargs={})

    def f_add_result(self,*args,**kwargs):
        ''' Adds a result under the current node.

        There are two ways to add a new result either by adding a result instance,
        directly:

        >>> new_result = Result('group1.group2.myresult', 1666, x=3, y=4, comment='Example!')
        >>> traj.f_add_result(new_result)

        Or by passing the values directly to the function, with the name being the first
        (non-keyword!) argument:

        >>> traj.f_add_result('group1.group2.myresult', 1666, x=3, y=3,comment='Example!')


        If you want to create a different result than the standard result, you can
        give the constructor as the first (non-keyword!) argument followed by the name (non-keyword!):

        >>> traj.f_add_result(PickleResult,'group1.group2.myresult', 1666, x=3, y=3, comment='Example!')

        Additional arguments (here `1666`) or keyword arguments (here `x=3, y=3`) are passed
        onto the constructor of the result.


        Adds the full name of the current node
        as prefix to the name of the result.
        If current node is the trajectory (root) adds the prefix `'results.trajectory'` to the full name.
        If current node is a single run (root) adds the prefix `'results.run_08%d%'` to the full name
        where `'08%d'` is replaced by the index of the current run.

        '''
        return self._nn_interface._add_generic(self,type_name = RESULT,
                                               group_type_name = RESULT_GROUP,
                                               args=args,kwargs=kwargs)


class DerivedParameterGroup(NNGroupNode):
    ''' Group node in your trajectory, hanging below `traj.derived_parameters`.

    You can add other groups or parameters to it.
    '''
    def f_add_derived_parameter_group(self,name):
        '''Adds an empty derived parameter group under the current node.

        Adds the full name of the current node
        as prefix to the name of the group.
        If current node is the trajectory (root) adds the prefix `'derived_parameters.trajectory'`
        to the full name.
        If current node is a single run (root) adds the prefix `'derived_parameters.run_08%d%'`
        to the full name
        where `'08%d'` is replaced by the index of the current run.

        The `name` can also contain subgroups separated via colons, for example:
        `name=subgroup1.subgroup2.subgroup3`. These other parent groups will be automatically
        be created.

        '''

        return self._nn_interface._add_generic(self,type_name = DERIVED_PARAMETER_GROUP,
                                               group_type_name = DERIVED_PARAMETER_GROUP,
                                               args = (name,), kwargs={})

    def f_add_derived_parameter(self,*args,**kwargs):
        ''' Adds a derived parameter under the current group.

        Similar to
        :func:`~pypet.naturalnaming.ParameterGroup.f_add_parameter`.

        Naming prefixes are added as in
        :func:`~pypet.naturalnaming.DerivedParameterGroup.f_add_derived_parameter_group`
        '''
        return self._nn_interface._add_generic(self,type_name = DERIVED_PARAMETER,
                                               group_type_name = DERIVED_PARAMETER_GROUP,
                                               args=args,kwargs=kwargs)



class ConfigGroup(NNGroupNode):
    ''' Group node in your trajectory, hanging below `traj.config`.

    You can add other groups or parameters to it.
    '''
    def f_add_config_group(self,name):
        '''Adds an empty config group under the current node.

        Adds the full name of the current node
        as prefix to the name of the group.
        If current node is the trajectory or single run (root), the prefix `'config'`
        is added to the full name.

        The `name` can also contain subgroups separated via colons, for example:
        `name=subgroup1.subgroup2.subgroup3`. These other parent groups will be automatically
        be created.

        '''
        return self._nn_interface._add_generic(self,type_name = CONFIG_GROUP,
                                               group_type_name = CONFIG_GROUP,
                                               args = (name,), kwargs={})

    def f_add_config(self,*args,**kwargs):
        ''' Adds config under the current group.

        Similar to
        :func:`~pypet.naturalnaming.ParameterGroup.f_add_parameter`.

        If current group is the trajectory the prefix `'config'` is added to the name.

        '''
        return self._nn_interface._add_generic(self,type_name = CONFIG,
                                               group_type_name = CONFIG_GROUP,
                                               args=args,kwargs=kwargs)







__author__ = 'Robert Meyer'

import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


verstr = "unknown"

# For versioning, Version found in pypet._version.py
try:
    verstrline = open('pypet/_version.py', "rt").read()
except EnvironmentError:
    pass # Okay, there is no version file.
else:
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise RuntimeError('Unable to find version in pypet/_version.py')


setup(
    name='pypet',
    version=verstr,
    packages=['pypet',
              'pypet.brian',
              'pypet.tests',
              'pypet.utils'],
    license='BSD',
    author='Robert Meyer',
    author_email='robert.meyer@ni.tu-berlin.de',
    description='A toolkit for numerical simulations to allow easy parameter exploration and storage of results.',
    long_description=open('long_description.txt').read(),
    url='https://github.com/SmokinCaterpillar/pypet',
    install_requires=[
        'tables >= 3.0.0',
        'pandas >= 0.12.0',
        'numpy >= 1.5.0'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Bio-Informatics']
)


__author__ = 'Robert Meyer'




''' Module for data containers, these are containers for parameters and results.
==========================================
General idea behind parameters and results
==========================================

The classes in this module are general containers for data that you acquire or need
to run any sort of scientific simulation in Python.

-----------------------------------------
Parameters
-----------------------------------------

A specific container is the so called parameter container (Base API is found in BaseParameter
class) it is used to keep data that is explicitly required as parameters for your simulations.
For example, this could be the numbers or cars for traffic jam simulations, or for
brain simulations the number of neurons, conductances of ion channels etc.

Parameter containers fulfill four further important
jobs:

 *  A key concept in numerical simulations is exploration of the parameter space. Therefore,
    the parameter containers not only keep a single value but can hold an exploration array
    of values.
    To avoid confusion with numpy arrays, I will explicitly always say exploration array.
    The individual values in the exploration array can be accessed one after the other
    for distinct simulations.

 *  It can be locked, meaning as soon as the parameter is assigned to hold a specific
    value and the value has already been used somewhere (for instance, creating X neurons),
    it cannot be changed any longer (except after being explicitly unlocked).
    This prevents the nasty error of having a particular parameter value
    at the beginning of a simulation but changing it during runtime for whatever reason. This
    can make your simulations impossible to understand by other people running them.
    Or even you running it again with different parameter settings but observing the
    very same results, since your parameter is changed later on in your simulations!
    In fact I ran into this problem during my PhD using someone else's simulations.
    Thus, debugging took ages. As a consequence this project was born.

    By definition parameters are fixed values, that once used never change.
    An exception to this rule is solely the exploration of the parameter space, but this
    requires to run a number of distinct simulations anyway.

 *  It provides a method to compare values in a strict boolean sense. For instance, this is
    important given you want to merge trajectories and delete simulations that have been done
    twice. The problem with most objects using them as parameters is that they have custom
    equality functions. Comparing two numpy arrays with '==' will returned a third array
    of truth values, for instance. But in order to identify equal parameter values,
    we need a single boolean value. Here the paramter container for numpy arrays could define
    the equality of two numpy arrays (with numpy as np) as
    >>> np.all(numpyarray1 == numpyarray2)
    which returns a single boolean value.

 *  It has a rough idea of how to serialize and store parameter values.
    If you intend to use basic parameter values like python int,str,bool or numpy arrays and
    data, you can ignore this part, otherwise go on:

    Conceptually it is always a good idea to have the storage of stuff and its use distinct. This
    allows you to change the storage method or service later on, to migrate from hdf5 to SQL,
    for example. This is also the case in this project. Yet, parameters know how to preprocess their
    values in order to serialize them. More precisley, they know how to turn their values
    into very basic data structures.
    They can convert values and their exploration arrays of values into stuff like
    numpy data types and numpy arrays, python native data types (int,str, etc) or pandas tables.
    These structures are then used by the independent storage service to actually put the data
    to disk (whatever that is).
    If a parameter is loaded from disk, these sort of basic data formats are given back
    to the parameter container which reconstructs the original value.

    Why is this important anyway?
    This becomes important as soon as you want to implement your own special parameter, that can
    use more complex objects as parameters. For instance, you have not only the number of cars
    as a parameter but a specific car object, that is principally complex in itself.
    This example is kind of far fetched, because the best idea would be to have the indivdual
    cars attributes as basic parameters (like number of doors, number of wheels etc.) and
    having the car construction as part of your simulation. But in pricniple this is possible.
    More realistic examples are BRIAN parameters, for the BRIAN neuron simulation package.
    A BRIAN value f_contains a number as well as a unit, so in order to store it to disk,
    both need to be separated.

Types of parameters supported so far:

 *  Parameter:
    Container for native python data: int,float,str,bool,complex
    Numpy data and arrays: np.int8-64, np.uint8-64, np.float32-64, np.complex, np.str
        Numpy arrays are allowed as well, however for larger numpy arrays, the ArrayParameter
        is recommended because it will keep a large array only once, even if it is used several
        times during exploration in the exploration array.

 *  ArrayParameter:
    Container for native python data as well as tuples and numpy arrays.
    The array parameter is the method of choice for large numpy arrays or python tuples.
    These are kept only once (and by the hdf5 storage service stored only once to disk)
    and in the exploration array you can find references to these arrays. This is particularly
    usefule if you reuse an array many times in distinct simulation, for example by exploring
    the parameter space in form of a cartesian product.


 *  PickleParameter:
    Container for all the data that can be pickled. Like the array parameter, distinct objects
    are kept only once and are refferred to in the exploration array


There is also the BrianParameter in the pypet.brian.parameter module.



------------------------------------
Results
------------------------------------

So far not many types of results (Base API found in BaseResult) exist
(only the Result and the BrianMonitorResult).
They provide less functionality than parameters. They can simply hold some simulation results.
For example, time course of traffic jam, voltage traces of neurons, etc.
If results are handed to these containers and they are in the right format, they
will in the end automatically be stored to disk.

Types of results supported so far:
 *  Result
    Container for basic data like native python types, numpy arrays and pandas tables (!).

 *  PickleResult
    Container for everything that can be pickled


There is also the BrianMonitorResult in the myper.brian.parameter module.

-----------------------------------
Object Table
-----------------------------------

Finally the ObjectTable class is a wrapper for pandas_ data frames.

.. _pandas: http://pandas.pydata.org/
'''
import logging
import petexceptions as pex
import numpy as np
from pypet.utils.helpful_functions import nested_equal, copydoc
from pypet import globally
from pandas import DataFrame
from pypet.naturalnaming import NNLeafNode

try:
    import cPickle as pickle
except:
    import pickle





class ObjectTable(DataFrame):
    ''' Wrapper class for pandas data frames.
    It creates data frames with dtype=object.

    Data stored into an object table preserves its original type. For instance, a python int
    is not automatically converted to a numpy 64 bit integer (np.int64).

    The object table serves as a standard data structure to hand data to a storage service.
    Given the default HDF5 storage service the types of the stored data are also preserved
    and reconstructed after loading from disk.

    '''
    def __init__(self, data=None, index = None, columns = None, copy=False):
        super(ObjectTable,self).__init__( data = data, index=index,columns=columns,
                                          dtype=object, copy=copy)





class BaseParameter(NNLeafNode):
    '''Abstract class that specifies the methods that need to be implemented for a trajectory
    parameter

    Parameters are simple container objects for data values. They handle single values as well as
    the so called exploration array. An array containing multiple values which are accessed
    one after the other in individual simulation runs.

    Parameter exploration is usually initiated through the trajectory see
    `:func:~pypet.trajectory.Trajectory.explore` and `:func:~pypet.trajectory.Trajectory.expand`.

    To access the parameter's data value one can call the :func:`f_get` method. Granted the parameter
    is explored via the trajectory, in order to
    access values of the exploration array, one first has to call the :func:`f_set_parameter_access`
    method with the index of the run and then use :func:`f_get`.

    Parameters support the concept of locking. Once a value of the parameter has been accessed,
    the parameter cannot be changed anymore unless it is explicitly unlocked using :func:`f_unlock`.
    This prevents parameters from being changed during runtime of a simulation.

    If multiprocessing is desired the parameter must be pickable!

    :param fullname: The fullname of the parameter in the trajectory tree, groupings are
        separated by a colon:
        `fullname = 'supergroup.subgroup.paramname'`

    :param comment: A useful comment describing the parameter:
        `comment = 'The number of cars for traffic jam simulations'`
    ''' 
    def __init__(self, full_name, comment=''):
        super(BaseParameter,self).__init__(full_name,comment, parameter=True)


        self._locked = False
        self._full_copy = False




    @property
    def v_locked(self):
        '''Whether or not the parameter is locked and prevents further modification'''
        return self._locked





    # @property
    # def v_value(self):
    #     '''The current value of the parameter'''
    #     return self.f_get()

    @property
    def v_full_copy(self):
        '''Whether or not the full parameter including the exploration array or only the current
        data is copied during pickling.

        If you run your simulations in multiprocessing mode, the whole trajectory and all
        parameters need to be pickled and are sent to the individual processes.
        Each process than runs an individual point in the parameter space trajectory.
        As a consequence, you do not need the exploration array during these calculations.
        Thus, if the full copy mode is f_set to False the parameter is pickled without
        the exploration array and you can save memory.

        If you want to access the full exploration array during individual runs, you need to set
        `v_full_copy` to True.

        It is recommended NOT to do that in order to save memory and also do obey the
        philosophy that individual simulations are independent.

        Example usage:

        >>> import pickle
        >>> param = Parameter('examples.fullcopy', data=333, comment='I show you how the copy mode works!')
        >>> param._explore([1,2,3,4])
        >>> dump=pickle.dumps(param)
        >>> newparam = pickle.loads(dump)
        >>> print newparam.f_get_array()
        ()
        >>> param.v_full_copy=True
        >>> dump = pickle.dumps(param)
        >>> newparam=pickle.loads(dump)
        >>> print newparam.f_get_array()
        (1,2,3,4)


        '''
        self._full_copy

    @v_full_copy.setter
    def v_full_copy(self,val):
        ''' Sets the full copy mode.'''
        val=bool(val)
        self._full_copy = val





    def f_is_array(self):
        ''' Returns true if the parameter is explored and contains an exploration array.
        '''
        raise NotImplementedError( "Should have implemented this." )

    def f_restore_default(self):
        ''' Restores original data if changed due to exploration.

        If a Parameter is explored, the actual data is changed over the course of different
        simulations. This method restores the original data assigned before exploration.

        '''
        raise NotImplementedError( "Should have implemented this." )



     
    def __len__(self):
        ''' Returns the length of the parameter.
        
        Only parameters that are explored can have a length larger than 1.
        If no values have been added to the parameter its length is 0.

        '''
        raise NotImplementedError( "Should have implemented this." )




    def f_val_to_str(self):
        ''' String summary of the value handled by the parameter.

        Note that representing
        the parameter as a string accesses its value, but for simpler debugging, this does not
        lock the parameter or counts as usage!

        String is truncated if it is longer or equal to the value specified in
        `:const:`globally.HDF5_STRCOL_MAX_COMMENT_LENGTH`

        '''
        old_locked = self._locked
        try :
            restr= str(self.f_get())

            if len(restr) >= globally.HDF5_STRCOL_MAX_VALUE_LENGTH:
                restr=restr[0:globally.HDF5_STRCOL_MAX_VALUE_LENGTH-3]+'...'

            return restr
        except Exception, e:
            return 'No Evaluation possible (yet)!'
        finally:
            self._locked = old_locked


    def f_supports(self, data):
        ''' Checks whether the data is supported by the parameter.

        '''
        return type(data) in globally.PARAMETER_SUPPORTED_DATA

    def _equal_values(self,val1,val2):
        ''' Checks if the parameter considers two values as equal.


        :return: True or False
        :raises: TypeError: If both values are not supported by the parameter.

        '''
        if self.f_supports(val1) != self.f_supports(val2):
            return False

        if not self.f_supports(val1) and not self.f_supports(val2):
                raise TypeError('I do not support the types of both inputs (>>%s<< and >>%s<<),'
                                ' therefore I cannot judge whether the two are equal.' %
                                str(type(val1)),str(type(val2)))

        if not self._values_of_same_type(val1,val2):
            return False

        return nested_equal(val1,val2)

    def _values_of_same_type(self,val1,val2):
        ''' Checks if two values agree in type.

        For example, two 32 bit integers would be of same type, but not a string and an integer,
        or not a 64 bit and a 32 bit integer.

        :return: True or False

        :raises: TypeError: if both values are not supported by the parameter.

        '''

        if self.f_supports(val1) != self.f_supports(val2):
            return False

        if not self.f_supports(val1) and not self.f_supports(val2):
                raise TypeError('I do not support the types of both inputs (>>%s<< and >>%s<<),'
                                ' therefore I cannot judge whether the two are of same type.' %
                                str(type(val1)),str(type(val2)))

        return type(val1) == type(val2)



    def __str__(self):
        if self.v_comment:

            returnstr = '<%s>: %s (Length:%d, Comment:%s): %s' % \
                        (self.f_get_class_name(), self.v_full_name, len(self),
                         self.v_comment, self.f_val_to_str())
        else:
            returnstr = '<%s>: %s (Length:%d): %s' % (self.f_get_class_name(),
                                                         self.v_full_name,
                                                         len(self), self.f_val_to_str())

        # if self.f_is_array():
        #     returnstr += ', Array: %s' %str(self.f_get_array())

        return returnstr


    def f_unlock(self):
        ''' Unlocks the locked parameter.

        Please use it very carefully, or best do not use this function at all. There should
        better be no reason to unlock a locked parameter!

        '''
        self._locked = False

    def f_lock(self):
        ''' Locks the parameter and forbids further manipulation.

        Changing the data value or exploration array of the parameter are no longer allowed.

        '''
        self._locked = True




    def f_set(self,data):
        ''' Sets specific values for a parameter.

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam', comment='I am a neat example')
        >>> param.f_set(44.0)
        >>> print parm.f_get()
        >>> 44.0

        :raises: ParameterLockedException:  if parameter is locked.

                 TypeError: If the parameter is an array or if the type of the
                                 data value is not supported by the parameter.

        '''
        raise NotImplementedError( "Should have implemented this." )

    def __getitem__(self, idx):
        '''  Equivalent to `f_get_array[idx]`

        :raises: TypeError if parameter is not an array
        '''
        return self.f_get_array().__getitem__(idx)


    def f_get(self):
        ''' Returns the current data value of the parameter and locks the parameter.

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam', comment='I am a neat example')
        >>> param.f_set(44.0)
        >>> print parm.f_get()
        >>> 44.0:

        '''

        raise NotImplementedError( "Should have implemented this." )

    def f_get_array(self):
        ''' Returns an iterable to iterate over the values of the exploration array.

        Note that the returned values should be either a copy of the exploration array
        or the array must be immutable, for example a python tuple.

        :return: immutable sequence

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam',data=22, comment='I am a neat example')
        >>> param._explore([42,43,43])
        >>> print param.f_get_array()
        >>> (42,43,44)

        '''

        raise NotImplementedError( "Should have implemented this." )
    
    def _explore(self, iterable):
        ''' The default method to create and explored parameter containing an array of entries.

        :param iterable: An iterable specifying the exploration array

                         For example:

                         >>> param = Parameter('groupA.groupB.myparam',data=22.33,\
                          comment='I am a neat example')
                         >>> param._explore([3.0,2.0,1.0])

        :raises: ParameterLockedExcpetion: if the parameter is locked.

                 TypeError: if the parameter is already an array.

        '''

        raise NotImplementedError( "Should have implemented this." )

    def _expand(self, iterable):
        ''' Similar to :func:`_explore` but appends to the exploration array.


        :param iterable: An iterable specifying the exploration array.


        :raises: ParameterLockedExcpetion: If the parameter is locked.

                 TypeError: if the parameter is not an array.

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam',data=3.13, comment='I am a neat example')
        >>> param._explore([3.0,2.0,1.0])
        >>> param._expand([42.0,43.0])
        >>> print param.f_get_array()
        >>> (3.0,2.0,1.0,42.0,43.0)

        '''
        raise NotImplementedError("Should have implemented this.")

    def f_set_parameter_access(self, idx=0):
        ''' Sets the current value according to the `idx` in the exploration array.

        Prepares the parameter for further usage, and tells it which point in the parameter
        space should be accessed by calls to :func:`~pypet.parameter.Parameter.f_get`.

        :param idx: The index within the exploration parameter

                    If the parameter is not an array, the single data value is considered
                    regardless of the value of `idx`.
                    Raises ValueError if the parameter is explored and `idx>=len(param)`

        :raises: ValueError: if the parameter is an array and `idx` is larger or equal to the
                             length of the parameter

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam',data=22.33, comment='I am a neat example')
        >>> param._explore([42.0,43.0,44.0])
        >>> param.f_set_parameter_access(idx=1)
        >>> print param.f_get()
        >>> 43.0

        '''
        raise NotImplementedError( "Should have implemented this." )
        
    def f_get_class_name(self):
        ''' Returns the name of the class i.e.
        `return self.__class__.__name__`

        '''
        return self.__class__.__name__


    def f_is_empty(self):
        ''' True if no data has been assigned to the parameter.

        >>> param = Parameter('myname.is.example', comment='I am _empty!')
        >>> param.f_is_empty()
        >>> True
        >>> param.f_set(444)
        >>> param.f_is_empty()
        >>> False

        '''
        return len(self) == 0

    def _shrink(self):
        ''' If a parameter is explored, i.e. it is an array, the whole exploration is deleted.

        Afterwards the parameter is no longer an array.

        Note that this function does not erase data from disk. So if the parameter has
        been stored with a service to disk and is shrunk, it can be restored by loading from
        disk.

        :raises: ParameterLockedException: if the parameter is locked.

                 TypeError: if  is the parameter is not an array.
        '''
        raise NotImplementedError( "Should have implemented this." )


    def f_empty(self):
        '''Erases all data in the parameter.

        Does not erase data from disk. So if the parameter has
        been stored with a service to disk and is emptied, it can be restored by loading from
        disk.

        :raises: ParameterLockedException: If the parameter is locked.

        '''
        raise NotImplementedError( "Should have implemented this." )






      
class Parameter(BaseParameter):
    ''' The standard parameter that handles access to simulation parameters.

    Parameters are simple container objects for data values. They handle single values as well as
    the so called exploration array. An array containing multiple values which is accessed
    one after the other in individual simulation runs.


    Parameter exploration is usually initiated through the trajectory see
    `:func:~pypet.trajectory.Trajectory.explore` and `:func:~pypet.trajectory.Trajectory.expand`.

    To access the parameter's data value one can call the :func:`~pypet.parameter.Parameter.f_get`
    method. Granted the parameter
    is explored via the trajectory, in order to
    access values of the exploration array, one first has to call the
    :func:`~pypet.parameter.Parameter.f_set_parameter_access`
    method with the index of the run and then use :func:`~pypet.parameter.Parameter.f_get`.

    Parameters support the concept of locking. Once a value of the parameter has been accessed,
    the parameter cannot be changed anymore unless it is explicitly unlocked using
    :func:`~pypet.parameter.Parameter.f_unlock`.
    Locking prevents parameters from being changed during runtime of a simulation.
    
    Supported data values for the parameter are

    * python natives (int,str,bool,float,complex),

    * numpy natives, arrays and matrices of type np.int8-64, np.uint8-64, np.float32-64,
      np.complex, np.str

    * python homogeneous not nested  tuples
    
    Note that for larger numpy arrays it is recommended to use the ArrayParameter.


     Example usage:

     >>> param = Parameter('traffic.mobiles.ncars',data=42, comment='I am a neat example')

    :param full_name: The full name of the parameter. Grouping can be achieved by using colons.

    :param data: A data value that is handled by the parameter. It is checked whether the parameter
        :func:`~pypet.parameter.Parameter.f_supports` the data. If not an TypeError is thrown.
        If the parameter becomes explored, the data value is kept as a default. After
        simulation the default value can be retained by calling
        :func:`~pypet.parameter.Parameter.f_restore_default`.
        The data can be accessed as follows:

        >>> param.f_get()
        42

        To change the data after parameter creation one can call :func:`~pypet.parameter.Parameter.f_set`:

        >>> param.f_set(43)
        >>> print param.f_get()
        43

    :param comment: A useful comment describing the parameter.
        The comment can be changed later on using the 'v_comment' variable.

        >>> print param.v_comment
        >>> 'I am a neat example'

    :raises: AttributeError: If `data` is not supported by the parameter.

    '''
    def __init__(self, full_name, data=None, comment=''):
        super(Parameter,self).__init__(full_name,comment)
        self._data= None
        self._default = None #The default value, which is the same as Data,
        # but it is necessary to keep a reference to it to restore the original value
        # after exploration
        self._explored_data=tuple()#The Explored Data
        self._set_logger()

        if not data == None:
            self.f_set(data)


    def _set_logger(self):
        self._logger = logging.getLogger('pypet.parameter.Parameter=' + self.v_full_name)

    def f_restore_default(self):
        ''' Restores the default data, that was set with the `:func:`~pypet.parameter.Parameter.f_set`
        method (or at initialisation).

        If the parameter is explored during the runtime of a simulation,
        the actual value of the parameter is changed and taken from the exploration array.
        Calling :func:`~pypet.parameter.Parameter.f_restore_default` sets the parameter's value
        back to it's original value.

        Example usage:

        >>> param = Parameter('supergroup1.subgroup2.', data=44, comment='Im a comment!')
        >>> param._explore([1,2,3,4])
        >>> param.f_set_parameter_access(2)
        >>> print param.f_get()
        >>> 3
        >>> param.f_restore_default()
        >>> print param.f_get()
        >>> 44

        '''
        self._data = self._default

    def __len__(self):
        if self._data is None:
            return 0
        elif len(self._explored_data)>0:
            return len(self._explored_data)
        else:
            return 1

    @copydoc(BaseParameter.f_is_array)
    def f_is_array(self):
        return len(self._explored_data)>0
       
    def __getstate__(self):
        ''' Returns the actual state of the parameter for pickling.

        '''
        result = self.__dict__.copy()


        # If we don't need a full copy of the Parameter (because a single process needs
        # only access to a single point in the parameter space we can delete the rest
        if not self._full_copy :
            result['_explored_data'] = tuple()

        del result['_logger'] #pickling does not work with loggers
        return result




    def __setstate__(self, statedict):
        ''' Sets the state for unpickling.

        '''
        self.__dict__.update( statedict)
        self._set_logger()
      
    @copydoc(BaseParameter.f_set_parameter_access)
    def f_set_parameter_access(self, idx=0):
        if idx >= len(self) and self.f_is_array():
            raise ValueError('You try to access the %dth parameter in the array of parameters, '
                             'yet there are only %d potential parameters.' % (idx, len(self)))
        else:
            self._data = self._explored_data[idx]

    def f_supports(self, data):
        ''' Checks if input data is supported by the parameter.'''

        if type(data) is tuple:
            for item in data:
                old_type = None
                if not type(item) in globally.PARAMETER_SUPPORTED_DATA:
                    return False
                if not old_type is None and old_type != type(item):
                    return False
                old_type = type(item)
            return True

        if type(data) in [np.ndarray, np.matrix]:
            dtype = data.dtype
            if np.issubdtype(dtype,np.str):
                dtype = np.str
        else:
            dtype=type(data)

        return dtype in globally.PARAMETER_SUPPORTED_DATA


    def _values_of_same_type(self,val1, val2):
        ''' Checks if two values agree in type.

        Raises a TypeError if both values are not supported by the parameter.
        Returns False if only one of the two values is supported by the parameter.

        Example usage:

        >>>param._values_of_same_type(42,43)
        True

        >>>param._values_of_same_type(42,'43')
        False

        :raises: TypeError

        :return: True or False

        '''
        if self.f_supports(val1) != self.f_supports(val2):
            return False

        if not self.f_supports(val1) and not self.f_supports(val2):
                raise TypeError('I do not support the types of both inputs (>>%s<< and >>%s<<),'
                                ' therefore I cannot judge whether the two are of same type.' %
                                str(type(val1)),str(type(val2)))
        
        if not type(val1) is type(val2):
            return False
        
        if type(val1) is np.array:
            if not val1.dtype is val2.dtype:
                return False
            
            if not np.shape(val1)==np.shape(val2):
                return False

        if type(val1) is tuple:
            return type(val1[0]) is type(val2[0])
        
        return True




    @copydoc(BaseParameter.f_set)
    def f_set(self,data):

        if self.v_locked:
            raise pex.ParameterLockedException('Parameter >>' + self._name + '<< is locked!')


        if self.f_is_array():
            raise AttributeError('Your Parameter is an explored array can no longer change values!')


        val = self._convert_data(data)

        if not self.f_supports(val):
            raise TypeError('Unsupported data type: ' +str(type(val)))

        self._data= val
        self._default = self._data




    def _convert_data(self, val):
        ''' Converts data to be handled by the parameter

        * sets numpy arrays immutable.

        :param val: the data value to convert

        :return: the converted data

        '''


        if type(val) is np.ndarray:
            val.flags.writeable = False
            return val

        return val


    @copydoc(BaseParameter.f_get_array)
    def f_get_array(self):
        ''' Returns an python tuple of the exploration array.


        :return: data tuple

        Example usage:

        >>> param = Parameter('groupA.groupB.myparam',data=22, comment='I am a neat example')
        >>> param._explore([42,43,43])
        >>> print param.f_get_array()
        >>> (42,43,44)

        '''
        if not self.f_is_array():
            raise TypeError('Your parameter is not array, so cannot return array')
        else:
            return self._explored_data


    def _explore(self, explore_iterable):
        ''' Explores the parameter according to the iterable.

        Raises ParameterLockedException if the parameter is locked.
        Raises TypeError if the parameter does not support the data,
        the types of the data in the iterable are not the same as the type of the default value,
        or the parameter has already an exploration array.

        Note that the parameter will iterate over the whole iterable once and store
        the individual data values into a tuple. Thus, the whole exploration array is
        explicitly stored in memory.

        :param iterable: An iterable specifying the exploration array

        For example:

        >>> param._explore([3.0,2.0,1.0])
        >>> param.f_get_array()
        (3.0,2.0,1.0)


        :raises TypeError,ParameterLockedException

        '''
        if self.v_locked:
            raise pex.ParameterLockedException('Parameter >>%s<< is locked!' % self.v_full_name)

        if self.f_is_array():
            raise TypeError('Your Parameter %s is already explored, cannot _explore it further!' % self._name)


        data_tuple = self._data_sanity_checks(explore_iterable)

        self._explored_data = data_tuple
        self.f_lock()

    def _expand(self,explore_iterable):
        ''' Explores the parameter according to the iterable and appends to the exploration array.

        Raises ParameterLockedException if the parameter is locked.
        Raises TypeError if the parameter does not support the data,
        the types of the data in the iterable are not the same as the type of the default value,
        or the parameter was not an array before.

        Note that the parameter will iterate over the whole iterable once and store
        the individual data values into a tuple. Thus, the whole exploration array is
        explicitly stored in memory.

        :param iterable: An iterable specifying the exploration array

                         For example:

                         >>> param = Parameter('Im.an.example', data=33.33, comment='Wooohoo!')
                         >>> param._explore([3.0,2.0,1.0])
                         >>> param.f_get_array()
                         >>> (3.0,2.0,1.0)

        :raises TypeError,ParameterLockedException

        '''
        if self.v_locked:
            raise pex.ParameterLockedException('Parameter >>%s<< is locked!' % self.v_full_name)

        if not self.f_is_array():
            raise TypeError('Your Parameter is not an array and can therefore not be expanded.' % self._name)


        data_tuple = self._data_sanity_checks(explore_iterable)



        self._explored_data = self._explored_data + data_tuple
        self.f_lock()


    def _data_sanity_checks(self, explore_iterable):
        ''' Checks if data values are legal.

        Checks if the data values are supported by the parameter and if the values are of the same
        type as the default value.
        '''
        data_tuple = []

        default_val = self._data

        for val in explore_iterable:
            newval = self._convert_data(val)


            if not self.f_supports(newval):
                raise TypeError('%s is of not supported type %s.' % (repr(val),str(type(newval))))

            if not self._values_of_same_type(newval,default_val):
                raise TypeError('Data is not of the same type as the original entry value, new type is %s vs old type %s.' % ( str(type(newval)),str(type(default_val))))


            data_tuple.append(newval)


        if len(data_tuple) == 0:
            raise ValueError('Cannot _explore an _empty list!')

        return tuple(data_tuple)

    @copydoc(BaseParameter._store)
    def _store(self):
        store_dict={}
        store_dict['data'] = ObjectTable(data={'data':[self._data]})
        if self.f_is_array():
            store_dict['explored_data'] = ObjectTable(data={'data':self._explored_data})


        return store_dict

    @copydoc(BaseParameter._load)
    def _load(self,load_dict):
        self._data = self._convert_data(load_dict['data']['data'][0])
        self._default=self._data
        if 'explored_data' in load_dict:
            self._explored_data = tuple([self._convert_data(x)
                                   for x in load_dict['explored_data']['data'].tolist()])



    @copydoc(BaseParameter.f_get)
    def f_get(self):
        self.f_lock() # As soon as someone accesses an entry the parameter gets locked
        return self._data

    @copydoc(BaseParameter._shrink)
    def _shrink(self):
        if self.f_is_empty():
            raise TypeError('Cannot _shrink _empty Parameter.')

        if self.v_locked:
            raise pex.ParameterLockedException('Parameter %s is locked!' % self.v_full_name)

        del self._explored_data
        self._explored_data={}

    @copydoc(BaseParameter.f_empty)
    def f_empty(self):
        if self.v_locked:
            raise pex.ParameterLockedException('Parameter %s is locked!' % self.v_full_name)

        self._shrink()
        del self._data
        self._data=None

     

class ArrayParameter(Parameter):

    ''' Similar to the :class:`:func:`~pypet.parameter.Parameter`, but recommended for
    large numpy arrays and python tuples.

    The array parameter is a bit smarter in memory management than the parameter.
    If a numpy array is used several times within an exploration, only one numpy array is stored by
    the default HDF5 storage service. For each individual run references to the corresponding
    numpy array are stored.

    '''

    IDENTIFIER = '__rr__'

    def _set_logger(self):
        self._logger = logging.getLogger('pypet.parameter.ArrayParameter=' + self.v_full_name)

    def _store(self):

        if not type(self._data) in [np.ndarray, tuple, np.matrix]:
            return super(ArrayParameter,self)._store()
        else:
            store_dict = {}

            store_dict['data'] = ObjectTable(columns=['data'+ArrayParameter.IDENTIFIER],index=[0])

            store_dict['data']['data'+ArrayParameter.IDENTIFIER] = 'data_array'


            store_dict['data_array'] = self._data

            if self.f_is_array():
                ## Supports smart storage by hashing numpy arrays are hashed by their data attribute
                smart_dict = {}

                store_dict['explored_data']=ObjectTable(columns=['data'+ArrayParameter.IDENTIFIER],index=range(len(self)))

                count = 0
                for idx,elem in enumerate(self._explored_data):

                    try:
                        hash(elem)
                        hash_elem = elem
                    except TypeError:
                        hash_elem = elem.data

                    if hash_elem in smart_dict:
                        name_id = smart_dict[hash_elem]
                        add = False
                    else:
                        name_id = count
                        add = True

                    name = self._build_name(name_id)
                    store_dict['explored_data']['data'+ArrayParameter.IDENTIFIER][idx] = name_id

                    if add:
                        store_dict[name] = elem
                        smart_dict[hash_elem] = name_id
                        count +=1

            return store_dict

    def _build_name(self,name_id):
        return 'ea_%s_%08d' % (ArrayParameter.IDENTIFIER,name_id)




    def _load(self,load_dict):
        data_table = load_dict['data']
        data_name = data_table.columns.tolist()[0]
        if ArrayParameter.IDENTIFIER in data_name:
            arrayname =  data_table['data'+ArrayParameter.IDENTIFIER][0]
            self._data = self._convert_data(load_dict[arrayname])


            if 'explored_data' in load_dict:
                explore_table = load_dict['explored_data']

                name_col = explore_table['data'+ArrayParameter.IDENTIFIER]

                explore_list = []
                for name_id in name_col:
                    arrayname = self._build_name(name_id)
                    explore_list.append(load_dict[arrayname])

                self._explored_data=tuple([self._convert_data(x) for x in explore_list])


        else:
            super(ArrayParameter,self)._load(load_dict)

        self._default=self._data





class PickleParameter(Parameter):
    ''' A parameter class that supports all pickable objects, and pickles everything!

    If you use the default HDF5 storage service, the pickle dumps are stored on disk.
    Works similar to the array parameter regarding memory management.

    '''
    IDENTIFIER='__pckl__'

    def _set_logger(self):
        self._logger = logging.getLogger('pypet.parameter.PickleParameter=' + self.v_full_name)

    def f_supports(self, data):
        ''' There is no straightforward check if an object can be pickled, so you have to take care that it can be pickled '''
        return True

    def _convert_data(self, val):
        return val

    def _build_name(self,name_id):
        return 'ed_%s_%08d' % (PickleParameter.IDENTIFIER,name_id)

    def _store(self):
        store_dict={}
        dump = pickle.dumps(self._data)
        store_dict['data']=ObjectTable(data={'data'+PickleParameter.IDENTIFIER : ['data_dump']})

        store_dict['data_dump'] = dump
        if self.f_is_array():

            store_dict['explored_data']=ObjectTable(columns=['data'+PickleParameter.IDENTIFIER],index=range(len(self)))

            smart_dict = {}
            count = 0


            for idx, val in enumerate(self._explored_data):

                obj_id = id(val)

                if obj_id in smart_dict:
                    name_id = smart_dict[obj_id]
                    add = False
                else:
                    name_id = count
                    add = True

                name = self._build_name(name_id)
                store_dict['explored_data']['data'+PickleParameter.IDENTIFIER][idx] = name_id

                if add:
                    store_dict[name] = pickle.dumps(val)
                    smart_dict[obj_id] = name_id
                    count +=1

        return store_dict

    def _load(self,load_dict):

        dump_name = load_dict['data']['data'+PickleParameter.IDENTIFIER][0]
        dump = load_dict[dump_name]
        self._data = pickle.loads(dump)

        if 'explored_data' in load_dict:
                explore_table = load_dict['explored_data']

                name_col = explore_table['data'+PickleParameter.IDENTIFIER]

                explore_list = []
                for name_id in name_col:
                    arrayname = self._build_name(name_id)
                    loaded = pickle.loads(load_dict[arrayname])
                    explore_list.append(loaded)

                self._explored_data=tuple(explore_list)


        self._default=self._data





      

            
                            

class BaseResult(NNLeafNode):
    ''' The basic api to store results.

    Compared to parameters (see :class: BaseParameter) results are also initialised with a fullname
    and a comment.
    As before grouping is achieved by colons in the name.

    Example usage:

    >>> result = Result(fullname='very.important.result',comment='I am important but f_emptyty :('

    '''
    def __init__(self, full_name, comment=''):
        super(BaseResult,self).__init__(full_name,comment, parameter=False)









class Result(BaseResult):
    ''' Light Container that stores tables and arrays.

    Note that no sanity checks on individual data is made
    and you have to take care, that your data is understood by the storage service.
    It is assumed that results tend to be large and therefore sanity checks would be too expensive.

    Data that can safely be stored into a Result are:

        * python natives (int,str,bool,float,complex),

        * numpy natives, arrays and matrices of type np.int8-64, np.uint8-64, np.float32-64,
          np.complex, np.str


        *

            python lists and tuples of the previous types
            (python natives + numpy natives and arrays)
            Lists and tuples are not allowed to be nested and must be
            homogeneous, i.e. only contain data of one particular type.
            Only integers, or only floats, etc.

        *
            python dictionaries of the previous types (not nested!), data can be
            heterogeneous, keys must be strings. For example, one key-value pair
            of string and int and one key-value pair of string and float, and so
            on.

        * pandas DataFrames_

        * :class:`pypet.parameter.ObjectTable`

    .. _DataFrames: http://pandas.pydata.org/pandas-docs/dev/dsintro.html#dataframe

    Such values are either set on initialisation of with :func:`~pypet.parameter.Result.f_set`

    Example usage:

    >>> res = Result('supergroup.subgroup.myresult', comment='I am a neat example!' \
        [1000,2000], {'a':'b','c':'d'}, hitchhiker='Arthur Dent')

    :param fullanme: The fullname of the result, grouping can be achieved are separated by colons,


    :param comment: A useful comment describing the parameter.
        The comment can later on be changed using the `v_comment` variable

        >>> param.v_comment
        'I am a neat example!'

    :param args: data that is handled by the result, it is kept by the result under the names
        `run%d` with `%d` being the position in the argument list.
        Can be changed or more can be added via :func:`~pypet.parameter.Result.f_set`

    :param kwargs: data that is handled by the result, it is kept by the result under the names
        specified by the keys of kwargs.
        Can be changed or more can be added via :func:`~pypet.parameter.Result.f_set`

        >>> print res.f_get(0)
        [1000,2000]
        >>> print res.f_get(1)
        {'a':'b','c':'d'}
        >>> print res.f_get('myresult')
        [1000,2000]
        >>> print res.f_get('hitchhiker')
        'ArthurDent'
        >>> print res.f_get('myresult','hitchhiker')
        ([1000,2000], 'ArthurDent')


    :raises: AttributeError: If the data in args or kwargs is not supported. Checks type of
                             outer data structure, i.e. checks if you have a list or dictionary.
                             But it does not check on individual values within dicts or lists.

    Alternatively one can also use :func:`~pypet.parameter.Result.f_set`

    >>> result.f_set('Uno',x='y')
    >>> print result.f_get(0)
    'Uno'
    >>> print result.f_get('x')
    'y'


    Alternative method to put and retrieve data from the result container is via `__getattr__` and
    `__setattr__`

    >>> res.ford = 'prefect'
    >>> res.ford
    'prefect'

    '''
    def __init__(self, full_name, *args, **kwargs):
        comment = kwargs.pop('comment','')
        super(Result,self).__init__(full_name,comment)
        self._data = {}
        self._set_logger()
        self.f_set(*args,**kwargs)
        self._no_data_string = False


    @property
    def v_no_data_string(self):
        '''Whether or not to give a short summarizing string when calling
         :func:`~pypet.parameter.Result.f_val_to_str`.

        Can be set to False if the evaluation of stored data into string is too costly.

        '''
        return self._no_data_string

    @v_no_data_string.setter
    def v_no_data_string(self,boolean):
        self._no_data_string(boolean)


    def __contains__(self, item):
        return item in self._data


    def f_val_to_str(self):
        ''' Summarizes data handled by the result as a string.
        Calls `__str__` on all handled data if `v_no_data_string=True`, else only
        the name/key of the handled data is printed.

        :return: string

        '''
        if not self._no_data_string:
            resstr=''
            for key,val in iter(sorted(self._data.items())):
                resstr+= '%s=%s, ' % (key, str(val))

                if len(resstr) >= globally.HDF5_STRCOL_MAX_VALUE_LENGTH:
                    resstr = resstr[0:globally.HDF5_STRCOL_MAX_VALUE_LENGTH-3]+'...'
                    return resstr

            resstr=resstr[0:-2]
            return resstr
        else:
            resstr = ', '.join(self._data.keys())

            if len(resstr) >= globally.HDF5_STRCOL_MAX_COMMENT_LENGTH:
                    resstr = resstr[0:globally.HDF5_STRCOL_MAX_COMMENT_LENGTH-3]+'...'

            return resstr



    def __str__(self):

        datastr = str([(key,str(type(val))) for key,val in self._data.iteritems()])
        if self.v_comment:
            return '<%s>: %s (Comment:%s): %s' % (self.f_get_class_name(),
                                                  self.v_full_name,self.v_comment,datastr)
        else:
            return '<%s>: %s: %s' % (self.f_get_class_name(),self.v_full_name,datastr)



    def _set_logger(self):
        self._logger = logging.getLogger('pypet.parameter.Result=' + self.v_full_name)


    def __getstate__(self):
        ''' Returns the actual state of the parameter for pickling.
        '''
        result = self.__dict__.copy()
        del result['_logger'] #pickling does not work with loggers
        return result


    def __setstate__(self, statedict):
        ''' Sets the state for unpickling.
        '''
        self.__dict__.update( statedict)
        self._set_logger()

    def f_to_dict(self, copy = True):
        ''' Returns all handled data as a dictionary

        :param copy: Whether the original dictionary or a shallow copy is returned. If you get
                     the real thing, please do not modify it!

        :return: Data dictionary

        '''
        if copy:
            return self._data.copy()
        else:
            return self._data


    def f_is_empty(self):
        ''' True if no data has been put into the result.

        Also True if all data has been erased via :func:`~pypet.parameter.Result.f_empty`

        '''
        return len(self._data)== 0

    @copydoc(BaseResult.f_empty)
    def f_empty(self):
        del self._data
        self._data={}

    def f_set(self,*args, **kwargs):
        ''' Method to put data into the result.

        :param args: The first positional argument is stored with the name of the result.
                    Following arguments are stored with `name_X` where `X` is the position
                    of the argument.

        :param kwargs: Arguments are stored with the key as name.

        >>> res = Result('supergroup.subgroup.myresult', comment='I am a neat example!')
        >>> res.f_set(333,42.0mystring='String!')
        >>> res.f_get('myresult')
        333
        >>> res.f_get('myresult_1')
        42.0
        >>> res.f_get(1)
        42.0
        >>> res.f_get('mystring')
        'String!'

        '''
        for idx,arg in enumerate(args):
            if idx == 0:
                valstr = self.v_name
            else:
                valstr = self.v_name+'_'+str(idx)
            self.f_set_single(valstr,arg)

        for key, arg in kwargs.items():
            self.f_set_single(key,arg)




    def f_get(self,*args):
        ''' Returns items handled by the result.

         If only a single name is given a single data item is returned, if several names are
         given, a list is returned. For integer inputs the results returns `resultname_X`.

         If the result contains only a single entry you can call `f_get()` without arguments.
         If you call `f_get()` and the result contains more than one element a ValueError is
         thrown.

        :param args: strings-names or integers

        :return: Single data item or tuple of data

        Example:

        >>> res = Result('supergroup.subgroup.myresult', comment='I am a neat example!' \
        [1000,2000], {'a':'b','c':'d'}, hitchhiker='Arthur Dent')
        >>> res.f_get('hitchhiker')
        'Arthur Dent'
        >>> res.f_get(0)
        [1000,2000]
        >>> res.f_get('hitchhiker', 'myresult')
        ('Arthur Dent', [1000,2000])

        '''

        if len(args) == 0:
            if len(self._data) == 1:
                return self._data[self._data.keys()[0]]
            elif len(self._data) >1:
                raise ValueError('Your result >>%s<< contains more than one entry: '
                                 '>>%s<< Please use >>f_get<< with one of these.' %
                                 (self.v_full_name, str(self._data.keys())))
            else:
                raise AttributeError('Your result >>%s<< is empty, cannot access data.' %
                                     self.v_full_name)

        result_list = []
        for name in args:
            if name == 0:
                name = self.v_name
            else:
                try:
                    name = self.v_name+'_%d' % name
                except TypeError:
                    pass

            if not name in self._data:
                raise  AttributeError('>>%s<< is not part of your result >>%s<<.' %
                                      (name,self.v_full_name))

            result_list.append(self._data[name])

        if len(args)==1:
            return result_list[0]
        else:
            return result_list

    def f_set_single(self, name, item):
        ''' Sets a single data item to the result.

        Raises AttributeError if the type of the data is not understood.
        Note that the type check is shallow. For example, if the data item is a list,
        the individual list elements are NOT checked whether their types are appropriate.

        :param name: The name of the data item
        :param item: The data item

        :raises: AttributeError

        Example usage:

        >>> res.f_set_single('answer', 42)
        >>> res.f_get('answer')
        >>> 42
        '''
        if name in ['comment', 'Comment']:
            self.v_comment= item

        if type(item) in ((np.ndarray,ObjectTable,DataFrame,dict,tuple,list,np.matrix)+
                             globally.PARAMETER_SUPPORTED_DATA):
            if (not type(item) in globally.PARAMETER_SUPPORTED_DATA) and len(item) == 0:
                self._logger.warning('The Item >>%s<< is _empty.' % name)

            self._data[name] = item
        else:
            raise AttributeError('Your result >>%s<< of type >>%s<< is not supported.' %
                                 (name,str(type(item))))



    @copydoc(NNLeafNode._store)
    def _store(self):
        store_dict ={}
        store_dict.update(self._data)
        return store_dict




    def _load(self, load_dict):
        self._data = load_dict



    def __delattr__(self, item):
        ''' Deletes an item from the result.

        If the item has been stored to disk before with a storage service, this storage is not
        deleted!

        :param item: The item to delete

        Example usage:

        >>> res = Result('Iam.an.example', comment = 'And a neat one, indeed!', fortytwo=42)
        >>> print 'fortytwo' in res
        >>> True
        >>> del res.fortytwo
        >>> print 'fortytwo' in res
        >>> False

        '''
        if item[0]=='_':
            del self.__dict__[item]
        elif item in self._data:
            del self._data[item]
        else:
            raise AttributeError('Your result >>%s<< does not contain %s.' % (self.name_,item))

    def __setattr__(self, key, value):
        if key[0]=='_':
            self.__dict__[key] = value
        elif hasattr(self.__class__,key):
            property = getattr(self.__class__,key)
            if property.fset is None:
                raise AttributeError('%s is read only!' % key)
            else:
                property.fset(self,value)
        else:
            self.f_set_single(key, value)

    def __getattr__(self, name):

        if (not '_data' in self.__dict__):

            raise AttributeError('This is to avoid pickling issues!')

        if name in ['Comment', 'comment']:
            return self._comment

        if not name in self._data:
            raise  AttributeError('>>%s<< is not part of your result >>%s<<.' % (name,self.v_full_name))

        return self._data[name]



class PickleResult(Result):
    ''' Result that digest everything and simply pickles it!

    Note that it is not checked whether data can be pickled, so take care that it works!
    '''


    def f_set_single(self, name, item):
        '''Adds a single data item to the pickle result.

         Note that it is NOT checked if the item can be pickled!

        '''
        self._data[name] = item


    def _set_logger(self):
        self._logger = logging.getLogger('pypet.parameter.PickleResult=' + self.v_full_name)



    def _store(self):
        store_dict ={}
        for key, val in self._data.items():
            store_dict[key] = pickle.dumps(val)
        return store_dict


    def _load(self, load_dict):
        for key, val in load_dict.items():
            self._data[key] = pickle.loads(val)


__author__ = 'Robert Meyer'


import unittest


from environment_test import *
from hdf5_storage_test import *
from parameter_test import *
from speed_test import *
from trajectory_test import *
from hdf5_merge_test import *
from hdf5_removal_and_continue_tests import *
from utilstest import *

# Works only if someone has installed Brian
try:
    from brian_parameter_test import *
    from brian_full_network_test import *
except ImportError:
    pass


if __name__ == '__main__':
    unittest.main()


'''
Created on 10.06.2013

@author: robert
'''


from pypet.parameter import Parameter,Result,ObjectTable
## Unfortunatley I have to this for read the docs to be able to mock brian
try:
    from brian.units import *
    from brian.stdunits import *
except TypeError:
    pass

from brian.fundamentalunits import Unit, Quantity, get_unit
from brian.monitor import SpikeMonitor,SpikeCounter,StateMonitor, \
    PopulationSpikeCounter, PopulationRateMonitor, StateSpikeMonitor,  \
    MultiStateMonitor, ISIHistogramMonitor, VanRossumMetric, Monitor

import numpy as np
import logging
import pandas as pd


class BrianParameter(Parameter):
    ''' A Parameter class that supports BRIAN Quantities.

    There are two storage modes:

    * :const:`~pypet.brian.parameter.BrianParameter.FLOAT_MODE`: ('FLOAT')

        The value is stored as a float and the unit as a sting.

        i.e. `12 mV` is stored as `12.0` and `'1.0 * mV'`

    * :const:`~pypet.brian.parameter.BrianParameter.STRING_MODE`: ('STRING')

        The value and unit are stored combined together as a string.

        i.e. `12 mV` is stored as `'12.0 * mV'`


    Supports data for the standard :class:`~pypet.parameter.Parameter`, too.

    '''

    IDENTIFIER = '__brn__'
    ''' Identification string stored into column title of hdf5 table'''

    FLOAT_MODE = 'FLOAT'
    '''Float storage mode'''
    STRING_MODE = 'STRING'
    '''String storage mode'''


    def __init__(self, full_name, data=None, comment='',storage_mode=FLOAT_MODE):
        super(BrianParameter,self).__init__(full_name,data,comment)

        self._storage_mode=None
        self.v_storage_mode = storage_mode

    @property
    def v_storage_mode(self):
        '''
        There are two storage modes:


        * :const:`~pypet.brian.parameter.BrianParameter.FLOAT_MODE`: ('FLOAT')

            The value is stored as a float and the unit as a sting.

            i.e. `12 mV` is stored as `12.0` and `'1.0 * mV'`

        * :const:`~pypet.brian.parameter.BrianParameter.STRING_MODE`: ('STRING')

            The value and unit are stored combined together as a string.

            i.e. `12 mV` is stored as `'12.0 * mV'`

        '''
        return self._storage_mode

    @v_storage_mode.setter
    def v_storage_mode(self, storage_mode):
        assert (storage_mode == BrianParameter.STRING_MODE or storage_mode == BrianParameter.FLOAT_MODE)
        self._storage_mode = storage_mode


    def _set_logger(self):
        self._logger = logging.getLogger('pypet.brian.parameter.BrianParameter=' + self.v_full_name)


    def f_supports(self, data):
        ''' Simply checks if data is supported '''
        if isinstance(data, Quantity):
            return True
        if super(BrianParameter,self).f_supports(data):
            return True
        return False

    def _values_of_same_type(self,val1, val2):

        if isinstance(val1,Quantity):
            try:
                if not val1.has_same_dimensions(val2):
                    return False
            except AttributeError:
                return False
        elif isinstance(val2,Quantity):
            try:
                if not val2.has_same_dimensions(val1):
                    return False
            except AttributeError:
                return False

        elif not super(BrianParameter,self)._values_of_same_type(val1, val2):
            return False


        return True

    def _store(self):

        if isinstance(self._data,Quantity):
            store_dict={}

            if self._storage_mode == BrianParameter.STRING_MODE:

                valstr = self._data.in_best_unit(python_code=True)
                store_dict['data'] = ObjectTable(data={'data'+BrianParameter.IDENTIFIER:[valstr],
                                                       'mode' +BrianParameter.IDENTIFIER:[self._storage_mode] })


                if self.f_is_array():
                    valstr_list = []
                    for val in self._explored_data:
                        valstr = val.in_best_unit(python_code=True)
                        valstr_list.append(valstr)


                    store_dict['explored_data'] = ObjectTable(data={'data'+BrianParameter.IDENTIFIER:valstr_list})

            elif self._storage_mode == BrianParameter.FLOAT_MODE:
                unitstr = repr(get_unit_fast(self._data))
                value = float(self._data)
                store_dict['data'] = ObjectTable(data={'value'+BrianParameter.IDENTIFIER:[value],
                                                       'unit'+BrianParameter.IDENTIFIER:[unitstr],
                                                       'mode' +BrianParameter.IDENTIFIER:[self._storage_mode]})

                if self.f_is_array():
                    value_list = []
                    for val in self._explored_data:
                        value = float(val)
                        value_list.append(value)


                    store_dict['explored_data'] = ObjectTable(data={'value'+BrianParameter.IDENTIFIER:value_list})

            else:
                raise RuntimeError('You shall not pass!')




            return store_dict
        else:
            return super(BrianParameter,self)._store()


    def _load(self,load_dict):
        data_table = load_dict['data']
        data_name = data_table.columns.tolist()[0]
        if BrianParameter.IDENTIFIER in data_name:
            self._storage_mode = data_table['mode'+BrianParameter.IDENTIFIER][0]
            if self._storage_mode == BrianParameter.STRING_MODE:
                valstr = data_table['data'+BrianParameter.IDENTIFIER][0]

                self._data = eval(valstr)


                if 'explored_data' in load_dict:
                    explore_table = load_dict['explored_data']

                    valstr_col = explore_table['data'+BrianParameter.IDENTIFIER]
                    explore_list = []
                    for valstr in valstr_col:
                        brian_quantity = eval(valstr)
                        explore_list.append(brian_quantity)

                    self._explored_data=tuple(explore_list)
            elif self._storage_mode == BrianParameter.FLOAT_MODE:

                # Recreate the brain units from the vale as float and unit as string:
                unit = eval(data_table['unit'+BrianParameter.IDENTIFIER][0])
                value = data_table['value'+BrianParameter.IDENTIFIER][0]
                self._data = value*unit

                if 'explored_data' in load_dict:
                    explore_table = load_dict['explored_data']

                    value_col = explore_table['value'+BrianParameter.IDENTIFIER]
                    explore_list = []
                    for value in value_col:
                        brian_quantity = value*unit
                        explore_list.append(brian_quantity)

                    self._explored_data=tuple(explore_list)


        else:
            super(BrianParameter,self)._load(load_dict)

        self._default = self._data



class BrianMonitorResult(Result):
    ''' A Result class that supports brian monitors.

    Monitor attributes are extracted and added as results with the attribute names.

    Add monitor on `__init__` via `monitor=` or via `f_set(monitor=brian_monitor)`

    Following monitors are supported:

    * SpikeCounter

    * VanRossumMetric

    * PopulationSpikeCounter

    * StateSpikeMonitor

    * PopulationRateMonitor

    * ISIHistogramMonitor:

    * SpikeMonitor

    *  MultiStateMonitor

    * StateMonitor


    Example:

    >>> brian_result = BrianMonitorResult('example.brian.mymonitor', monitor=SpikeMonitor(...), comment='Im a SpikeMonitor Example!')
    >>> brian_result.nspikes
    1337


    There are two storage modes in case you use the SpikeMonitor and StateSpikeMonitor:

    * :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`: ('TABLE')

        Default, information is stored into a single table where
        the first column is the neuron index,
        second column is the spike time
        following columns contain variable values (in case of the StateSpikeMonitor)
        This is a very compact storage form.

    * :const:`~pypet.brian.parameter.BrianParameter.ARRAY_MODE`: ('ARRAY')

        For each neuron there will be a new hdf5 array,
        i.e. if you have many neurons your result node will have many entries.
        Note that this mode does sort everything according to the neurons but
        reading and writing of data might take muuuuuch longer compared to
        the other mode.

    '''

    TABLE_MODE = 'TABLE'
    '''Table storage mode for SpikeMonitor and StateSpikeMonitor'''
    ARRAY_MODE = 'ARRAY'
    '''Array storage mode, not recommended if you have many neurons!'''

    #keywords=set(['data','values','spikes','times','rate','count','mean_var',])

    def __init__(self, full_name, *args, **kwargs):
        super(BrianMonitorResult,self).__init__(full_name)

        self._storage_mode=None
        storage_mode = kwargs.pop('storage_mode',BrianMonitorResult.TABLE_MODE)
        self.v_storage_mode=storage_mode

        self.f_set(*args,**kwargs)


    @property
    def v_storage_mode(self):
        '''The storage mode for SpikeMonitor and StateSpikeMonitor

        There are two storage modes:

        * :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`: ('TABLE')

            Default, information is stored into a single table where
            the first column is the neuron index,
            second column is the spike time
            following columns contain variable values (in case of the StateSpikeMonitor)
            This is a very compact storage form.

        * :const:`~pypet.brian.parameter.BrianParameter.ARRAY_MODE`: ('ARRAY')

            For each neuron there will be a new hdf5 array,
            i.e. if you have many neurons your result node will have many entries.
            Note that this mode does sort everything according to the neurons but
            reading and writing of data might take muuuuuch longer compared to
            the other mode.

        '''
        return self._storage_mode

    @v_storage_mode.setter
    def v_storage_mode(self, storage_mode):
        assert (storage_mode == BrianMonitorResult.ARRAY_MODE or storage_mode == BrianMonitorResult.TABLE_MODE)
        self._storage_mode = storage_mode

    def f_set_single(self, name, item):
        ''' To add a monitor use `f_set_single('monitor', brian_monitor)`.

        Otherwise `f_set_single` works similar to :func:`~pypet.parameter.Result.f_set_single`.
        '''

        if name == 'storage_mode':
            self.v_storage_mode=item

        elif isinstance(item, (Monitor, MultiStateMonitor)):
            self._extract_monitor_data(item)
        else:
            super(BrianMonitorResult,self).f_set_single(name,item)


    
    def _extract_monitor_data(self,monitor):
        ## Check for each monitor separately:



        if isinstance(monitor, SpikeCounter):
            self._extract_spike_counter(monitor)

        elif isinstance(monitor, VanRossumMetric):
            self._extract_van_rossum_metric(monitor)

        elif isinstance(monitor, PopulationSpikeCounter):
            self._extract_population_spike_counter(monitor)

        elif isinstance(monitor, StateSpikeMonitor):
            self._extract_state_spike_monitor(monitor)
        
        elif  isinstance(monitor, PopulationRateMonitor):
            self._extract_population_rate_monitor(monitor)


        elif isinstance(monitor, ISIHistogramMonitor):
            self._extract_isi_hist_monitor(monitor)

        elif isinstance(monitor,SpikeMonitor):
            self._extrac_spike_monitor(monitor)

        elif isinstance(monitor, MultiStateMonitor):
            self._extract_multi_state_monitor(monitor)
        
        elif isinstance(monitor,StateMonitor):
            self._extract_state_monitor(monitor)

            
        else:
            raise ValueError('Monitor Type %s is not supported (yet)' % str(type(monitor)))

    def _extract_spike_counter(self,monitor):
        self.f_set(count=monitor.count)
        self.f_set(source = str(monitor.source))
        self.f_set(delay = monitor.delay)
        self.f_set(nspikes = monitor.nspikes)


    def _extract_van_rossum_metric(self,monitor):

        self.f_set(source = str(monitor.source))
        self.f_set(N=monitor.N)
        self.f_set(tau=float(monitor.tau))
        self.f_set(tau_unit = 'second')
        #self.f_set(timestep = monitor.timestep)

        self.f_set(distance = monitor.distance)



    def _extract_isi_hist_monitor(self,monitor):

        self.f_set(source = str(monitor.source))
        self.f_set(count = monitor.count)
        self.f_set(bins = monitor.bins)
        self.f_set(delay = monitor.delay)
        self.f_set(nspikes = monitor.nspikes)



    def _extract_state_spike_monitor(self,monitor):

        self.f_set(source = str(monitor.source))



        varnames = monitor._varnames
        if not isinstance(varnames, tuple) :
            varnames = (varnames,)

        self.f_set(varnames = varnames)

        #self.f_set(record = monitor.record)
        self.f_set(delay=monitor.delay)
        self.f_set(nspikes = monitor.nspikes)

        # if hasattr(monitor, 'function'):
        #     data_dict['function'] = [getsource(monitor.function)]

        self.f_set(nspikes = monitor.nspikes)
        self.f_set(times_unit = 'second')

        if self._storage_mode==BrianMonitorResult.TABLE_MODE:
            spike_dict={}

            if len(monitor.spikes)>0:
                zip_lists = zip(*monitor.spikes)
                time_list = zip_lists[1]

                nounit_list = [np.float64(time) for time in time_list]

                spike_dict['times'] = nounit_list
                spike_dict['neuron'] = list(zip_lists[0])

                count = 2
                for varname in varnames:

                    var_list = list(zip_lists[count])
                    self.f_set(**{varname+'_unit':  repr(get_unit(var_list[0]))})
                    nounit_list = [np.float64(var) for var in var_list]
                    spike_dict[varname] = nounit_list
                    count = count +1

                self.f_set(spikes=pd.DataFrame(data=spike_dict))

        elif self._storage_mode==BrianMonitorResult.ARRAY_MODE:
                for neuron in range(len(monitor.source)):
                    spikes = monitor.times(neuron)
                    if len(spikes)>0:
                        key = 'spiketimes_%08d' % neuron
                        self.f_set(**{key:spikes})

                for varname in varnames:
                     for neuron in range(len(monitor.source)):
                         values = monitor.values(varname,neuron)
                         if len(values)>0:
                             key = varname+'_unit'
                             if not  key in self:
                                 self.f_set(**{key:  repr(get_unit(values[0]))})
                             key = varname+'_idx%08d' % neuron
                             self.f_set(**{key:values})


        else:
                raise RuntimeError('You shall not pass!')


     
    def _extrac_spike_monitor(self,monitor):
        
        #assert isinstance(monitor, SpikeMonitor)


        self.f_set(source = str(monitor.source))

        self.f_set(record = monitor.record)

        self.f_set(nspikes = monitor.nspikes)

        self.f_set(times_unit='second')

        self.f_set(delay=monitor.delay)


        if self._storage_mode==BrianMonitorResult.TABLE_MODE:
            spike_dict={}

            if len(monitor.spikes)>0:
                zip_lists = zip(*monitor.spikes)
                time_list = zip_lists[1]

                nounit_list = [np.float64(time) for time in time_list]

                spike_dict['times'] = nounit_list
                spike_dict['neuron'] = list(zip_lists[0])

                spikeframe = pd.DataFrame(data=spike_dict)
                self.f_set(spikes=spikeframe)

        elif self._storage_mode==BrianMonitorResult.ARRAY_MODE:
                for neuron, spikes in monitor.spiketimes.items():
                    if len(spikes)>0:
                        key = 'spiketimes_%08d' % neuron
                        self.f_set(**{key:spikes})

        else:
                raise RuntimeError('You shall not pass!')



    def _extract_population_rate_monitor(self,monitor):
        assert isinstance(monitor, PopulationRateMonitor)
        

        self.f_set(source = str(monitor.source))
        self.f_set(times_unit = 'second', rate_unit='Hz')
        self.f_set(times = monitor.times)
        self.f_set(rate = monitor.rate)
        self.f_set(delay = monitor.delay)
        self.f_set(bin=monitor._bin)
        

    def _extract_spike_counter(self,monitor):

        self.f_set(nspikes = monitor.nspikes)
        self.f_set(source = str(monitor.source))
        self.f_set(count=monitor.count)
        self.f_set(delay=monitor.delay)



    def _extract_population_spike_counter(self,monitor):
        
        assert isinstance(monitor, PopulationSpikeCounter)

        self.f_set(nspikes = monitor.nspikes)
        self.f_set(source = str(monitor.source))
        self.f_set(delay=monitor.delay)




    def _extract_multi_state_monitor(self,monitors):

        self.f_set(vars = monitors.vars)


        if len(monitors.times)>0:
            self.f_set(times = monitors.times)
            self.f_set(times_unit = 'second')

        ### Store recorded values ###
        for idx,varname in enumerate(monitors.vars):
            monitor = monitors.monitors[varname]
            if idx == 0:


                self.f_set(record = monitor.record)

                self.f_set(when = monitor.when)

                self.f_set(timestep = monitor.timestep)

                self.f_set(source = str(monitor.P))


            self.f_set(**{varname+'_mean':monitor.mean})
            self.f_set(**{varname+'_var' : monitor.var})
            if len(monitors.times)>0:
                self.f_set(**{varname+'_values' : monitor.values})
            self.f_set (**{varname+'_unit' : repr(monitor.unit)})


    def _extract_state_monitor(self,monitor):

        self.f_set(varname = monitor.varname)
        self.f_set(unit = repr(monitor.unit))

            
        self.f_set(record = monitor.record)
        
        self.f_set(when = monitor.when)
        
        self.f_set(timestep = monitor.timestep)

        self.f_set(source = str(monitor.P))
        

        self.f_set(mean = monitor.mean)
        self.f_set(var = monitor.var)
        if len(monitor.times)>0:
            self.f_set(times = monitor.times)
            self.f_set(values = monitor.values)
            self.f_set(times_unit = 'second')

            
